from schnetpack.md.calculators import MDCalculator
from schnetpack.md.utils import MDUnits
from schnetpack import Properties


class SGDMLCalculator(MDCalculator):
    """
    Calculator for the sGDML model published in [#sgdml1]_ and [#sgdml2]_ . This model predicts energies and forces and
    currently only is available for molecules of the same size. In order to use the calculator, the sgdml code package
    available online at https://github.com/stefanch/sGDML and described in [#sgdml3]_ is required.

    Args:
        model (torch.nn.module): Loaded sGDML model.
        required_properties (list): Properties to predict with sGDML, available are energies and forces, which are
                                    default.
        force_handle (str): Name of forces in output. Default is 'forces'.
        force_conversion (float): Convert forces from sGDML to atomic units. sGDML used kcal/mol/Angstrom, so this
                                  default is used here.
        position_conversion (float): Unit conversion for the length used in the model computing all properties. E.g. if
                             the model needs Angstrom, one has to provide the conversion factor converting from the
                             atomic units used internally (Bohr) to Angstrom: 0.529177. Since sGDML uses Angstrom and
                             the MD algorithm atomic units, this factor is used as default here
        property_conversion (dict(float)): Optional dictionary of conversion factors for other properties predicted by
                                           the model. Only changes the units used for logging the various outputs.
        detach (bool): Detach property computation graph after every calculator call. Enabled by default. Should only
                       be disabled if one wants to e.g. compute derivatives over short trajectory snippets.

    References
    ----------
    .. [#sgdml1] Chmiela, Tkatchenko, Sauceda, Poltavsky, Schütt, Müller:
       Energy-conserving Molecular Force Fields.
       Science Advances, 3 (5), e1603015. 2017.
    .. [#sgdml2] Chmiela, Sauceda, Müller, Tkatchenko:
       Towards Exact Molecular Dynamics Simulations with Machine-Learned Force Fields.
       Nature Communications, 9 (1), 3887. 2018.
    .. [#sgdml3] Chmiela, Sauceda, Poltavsky, Müller, Tkatchenko:
       sGDML: Constructing accurate and data efficient molecular force fields using machine learning.
       Computer Physics Communications (in press). https://doi.org/10.1016/j.cpc.2019.02.007
    """

    def __init__(
        self,
        model,
        required_properties=[Properties.energy, Properties.forces],
        force_handle=Properties.forces,
        position_conversion="Angstrom",
        force_conversion="kcal/mol/Angstrom",
        property_conversion={},
        detach=True,
    ):
        super(SGDMLCalculator, self).__init__(
            required_properties,
            force_handle,
            position_conversion,
            force_conversion,
            property_conversion,
            detach,
        )

        self.model = model

    def calculate(self, system):
        """
        Main routine, extract the positions from the current system, use the sGDML model to predict the energies and
        forces and construct the results dictionary, which is then used to update the system.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.
        """
        inputs = self._generate_input(system)
        energy, forces = self.model(inputs)

        self.results = {Properties.energy: energy, Properties.forces: forces}
        self._update_system(system)

    def _generate_input(self, system):
        """
        Function to extracts atom_types, positions and atom_masks from the system and generate a properly
        formatted input for sGDML, which in this case is just the array of positions.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.

        Returns:
            torch.Tensor: sGDML inputs, which are the positions in the format N_replicas*N_molecules x N_atoms x 3.
        """
        positions, atom_types, atom_masks, cells, pbc = self._get_system_molecules(
            system
        )

        return positions



import torch
from torch import nn

from schnetpack.nn.cutoff import CosineCutoff

__all__ = [
    "AngularDistribution",
    "BehlerAngular",
    "GaussianSmearing",
    "RadialDistribution",
]


class AngularDistribution(nn.Module):
    """
    Routine used to compute angular type symmetry functions between all atoms i-j-k, where i is the central atom.

    Args:
        radial_filter (callable): Function used to expand distances (e.g. Gaussians)
        angular_filter (callable): Function used to expand angles between triples of atoms (e.g. BehlerAngular)
        cutoff_functions (callable): Cutoff function
        crossterms (bool): Include radial contributions of the distances r_jk
        pairwise_elements (bool): Recombine elemental embedding vectors via an outer product. If e.g. one-hot encoding
            is used for the elements, this is equivalent to standard Behler functions
            (default=False).

    """

    def __init__(
        self,
        radial_filter,
        angular_filter,
        cutoff_functions=CosineCutoff,
        crossterms=False,
        pairwise_elements=False,
    ):
        super(AngularDistribution, self).__init__()
        self.radial_filter = radial_filter
        self.angular_filter = angular_filter
        self.cutoff_function = cutoff_functions
        self.crossterms = crossterms
        self.pairwise_elements = pairwise_elements

    def forward(self, r_ij, r_ik, r_jk, triple_masks=None, elemental_weights=None):
        """
        Args:
            r_ij (torch.Tensor): Distances to neighbor j
            r_ik (torch.Tensor): Distances to neighbor k
            r_jk (torch.Tensor): Distances between neighbor j and k
            triple_masks (torch.Tensor): Tensor mask for non-counted pairs (e.g. due to cutoff)
            elemental_weights (tuple of two torch.Tensor): Weighting functions for neighboring elements, first is for
                                                            neighbors j, second for k

        Returns:
            torch.Tensor: Angular distribution functions

        """

        nbatch, natoms, npairs = r_ij.size()

        # compute gaussilizated distances and cutoffs to neighbor atoms
        radial_ij = self.radial_filter(r_ij)
        radial_ik = self.radial_filter(r_ik)
        angular_distribution = radial_ij * radial_ik

        if self.crossterms:
            radial_jk = self.radial_filter(r_jk)
            angular_distribution = angular_distribution * radial_jk

        # Use cosine rule to compute cos( theta_ijk )
        cos_theta = (torch.pow(r_ij, 2) + torch.pow(r_ik, 2) - torch.pow(r_jk, 2)) / (
            2.0 * r_ij * r_ik
        )

        # Required in order to catch NaNs during backprop
        if triple_masks is not None:
            cos_theta[triple_masks == 0] = 0.0

        angular_term = self.angular_filter(cos_theta)

        if self.cutoff_function is not None:
            cutoff_ij = self.cutoff_function(r_ij).unsqueeze(-1)
            cutoff_ik = self.cutoff_function(r_ik).unsqueeze(-1)
            angular_distribution = angular_distribution * cutoff_ij * cutoff_ik

            if self.crossterms:
                cutoff_jk = self.cutoff_function(r_jk).unsqueeze(-1)
                angular_distribution = angular_distribution * cutoff_jk

        # Compute radial part of descriptor
        if triple_masks is not None:
            # Filter out nan divisions via boolean mask, since
            # angular_term = angular_term * triple_masks
            # is not working (nan*0 = nan)
            angular_term[triple_masks == 0] = 0.0
            angular_distribution[triple_masks == 0] = 0.0

        # Apply weights here, since dimension is still the same
        if elemental_weights is not None:
            if not self.pairwise_elements:
                Z_ij, Z_ik = elemental_weights
                Z_ijk = Z_ij * Z_ik
                angular_distribution = (
                    torch.unsqueeze(angular_distribution, -1)
                    * torch.unsqueeze(Z_ijk, -2).float()
                )
            else:
                # Outer product to emulate vanilla SF behavior
                Z_ij, Z_ik = elemental_weights
                B, A, N, E = Z_ij.size()
                pair_elements = Z_ij[:, :, :, :, None] * Z_ik[:, :, :, None, :]
                pair_elements = pair_elements + pair_elements.permute(0, 1, 2, 4, 3)
                # Filter out lower triangular components
                pair_filter = torch.triu(torch.ones(E, E)) == 1
                pair_elements = pair_elements[:, :, :, pair_filter]
                angular_distribution = torch.unsqueeze(
                    angular_distribution, -1
                ) * torch.unsqueeze(pair_elements, -2)

        # Dimension is (Nb x Nat x Nneighpair x Nrad) for angular_distribution and
        # (Nb x Nat x NNeigpair x Nang) for angular_term, where the latter dims are orthogonal
        # To multiply them:
        angular_distribution = (
            angular_distribution[:, :, :, :, None, :]
            * angular_term[:, :, :, None, :, None]
        )
        # For the sum over all contributions
        angular_distribution = torch.sum(angular_distribution, 2)
        # Finally, we flatten the last two dimensions
        angular_distribution = angular_distribution.view(nbatch, natoms, -1)

        return angular_distribution


class BehlerAngular(nn.Module):
    """
    Compute Behler type angular contribution of the angle spanned by three atoms:

    :math:`2^{(1-\zeta)} (1 + \lambda \cos( {\\theta}_{ijk} ) )^\zeta`

    Sets of zetas with lambdas of -1 and +1 are generated automatically.

    Args:
        zetas (set of int): Set of exponents used to compute angular Behler term (default={1})

    """

    def __init__(self, zetas={1}):
        super(BehlerAngular, self).__init__()
        self.zetas = zetas

    def forward(self, cos_theta):
        """
        Args:
            cos_theta (torch.Tensor): Cosines between all pairs of neighbors of the central atom.

        Returns:
            torch.Tensor: Tensor containing values of the angular filters.
        """
        angular_pos = [
            2 ** (1 - zeta) * ((1.0 - cos_theta) ** zeta).unsqueeze(-1)
            for zeta in self.zetas
        ]
        angular_neg = [
            2 ** (1 - zeta) * ((1.0 + cos_theta) ** zeta).unsqueeze(-1)
            for zeta in self.zetas
        ]
        angular_all = angular_pos + angular_neg
        return torch.cat(angular_all, -1)


def gaussian_smearing(distances, offset, widths, centered=False):
    r"""Smear interatomic distance values using Gaussian functions.

    Args:
        distances (torch.Tensor): interatomic distances of (N_b x N_at x N_nbh) shape.
        offset (torch.Tensor): offsets values of Gaussian functions.
        widths: width values of Gaussian functions.
        centered (bool, optional): If True, Gaussians are centered at the origin and
            the offsets are used to as their widths (used e.g. for angular functions).

    Returns:
        torch.Tensor: smeared distances (N_b x N_at x N_nbh x N_g).

    """
    if not centered:
        # compute width of Gaussian functions (using an overlap of 1 STDDEV)
        coeff = -0.5 / torch.pow(widths, 2)
        # Use advanced indexing to compute the individual components
        diff = distances[:, :, :, None] - offset[None, None, None, :]
    else:
        # if Gaussian functions are centered, use offsets to compute widths
        coeff = -0.5 / torch.pow(offset, 2)
        # if Gaussian functions are centered, no offset is subtracted
        diff = distances[:, :, :, None]
    # compute smear distance values
    gauss = torch.exp(coeff * torch.pow(diff, 2))
    return gauss


class GaussianSmearing(nn.Module):
    r"""Smear layer using a set of Gaussian functions.

    Args:
        start (float, optional): center of first Gaussian function, :math:`\mu_0`.
        stop (float, optional): center of last Gaussian function, :math:`\mu_{N_g}`
        n_gaussians (int, optional): total number of Gaussian functions, :math:`N_g`.
        centered (bool, optional): If True, Gaussians are centered at the origin and
            the offsets are used to as their widths (used e.g. for angular functions).
        trainable (bool, optional): If True, widths and offset of Gaussian functions
            are adjusted during training process.

    """

    def __init__(
        self, start=0.0, stop=5.0, n_gaussians=50, centered=False, trainable=False
    ):
        super(GaussianSmearing, self).__init__()
        # compute offset and width of Gaussian functions
        offset = torch.linspace(start, stop, n_gaussians)
        widths = torch.FloatTensor((offset[1] - offset[0]) * torch.ones_like(offset))
        if trainable:
            self.width = nn.Parameter(widths)
            self.offsets = nn.Parameter(offset)
        else:
            self.register_buffer("width", widths)
            self.register_buffer("offsets", offset)
        self.centered = centered

    def forward(self, distances):
        """Compute smeared-gaussian distance values.

        Args:
            distances (torch.Tensor): interatomic distance values of
                (N_b x N_at x N_nbh) shape.

        Returns:
            torch.Tensor: layer output of (N_b x N_at x N_nbh x N_g) shape.

        """
        return gaussian_smearing(
            distances, self.offsets, self.width, centered=self.centered
        )


class RadialDistribution(nn.Module):
    """
    Radial distribution function used e.g. to compute Behler type radial symmetry functions.

    Args:
        radial_filter (callable): Function used to expand distances (e.g. Gaussians)
        cutoff_function (callable): Cutoff function
    """

    def __init__(self, radial_filter, cutoff_function=CosineCutoff):
        super(RadialDistribution, self).__init__()
        self.radial_filter = radial_filter
        self.cutoff_function = cutoff_function

    def forward(self, r_ij, elemental_weights=None, neighbor_mask=None):
        """
        Args:
            r_ij (torch.Tensor): Interatomic distances
            elemental_weights (torch.Tensor): Element-specific weights for distance functions
            neighbor_mask (torch.Tensor): Mask to identify positions of neighboring atoms

        Returns:
            torch.Tensor: Nbatch x Natoms x Nfilter tensor containing radial distribution functions.
        """

        nbatch, natoms, nneigh = r_ij.size()

        radial_distribution = self.radial_filter(r_ij)

        # If requested, apply cutoff function
        if self.cutoff_function is not None:
            cutoffs = self.cutoff_function(r_ij)
            radial_distribution = radial_distribution * cutoffs.unsqueeze(-1)

        # Apply neighbor mask
        if neighbor_mask is not None:
            radial_distribution = radial_distribution * torch.unsqueeze(
                neighbor_mask, -1
            )

        # Weigh elements if requested
        if elemental_weights is not None:
            radial_distribution = (
                radial_distribution[:, :, :, :, None]
                * elemental_weights[:, :, :, None, :].float()
            )

        radial_distribution = torch.sum(radial_distribution, 2)
        radial_distribution = radial_distribution.view(nbatch, natoms, -1)
        return radial_distribution



import torch
import torch.nn as nn
from torch.autograd import grad

import schnetpack as spk
import schnetpack.nn as spknn
from schnetpack import Properties
from schnetpack.md.calculators import SchnetPackCalculator


class PairwiseRepresentation(nn.Module):
    def __init__(self):
        super(PairwiseRepresentation, self).__init__()
        self.distances = spknn.AtomDistances()

    def forward(self, inputs):
        positions = inputs[Properties.R]
        neighbors = inputs[Properties.neighbors]
        neighbor_mask = inputs[Properties.neighbor_mask]
        cell = inputs[Properties.cell]
        cell_offsets = inputs[Properties.cell_offset]

        distances = self.distances(
            positions,
            neighbors,
            cell=cell,
            cell_offsets=cell_offsets,
            neighbor_mask=neighbor_mask,
        )

        return distances


class LJAtomistic(nn.Module):
    def __init__(
        self,
        r_equilibrium=1.0,
        well_depth=10.0,
        cutoff=5.0,
        property="y",
        derivative=None,
        negative_dr=False,
        stress=None,
        create_graph=True,
    ):
        super(LJAtomistic, self).__init__()

        self.r_equilibrium = r_equilibrium
        self.well_depth = well_depth

        self.create_graph = create_graph
        self.property = property
        self.derivative = derivative
        self.negative_dr = negative_dr
        self.stress = stress

        self.cutoff = CustomCutoff(cutoff)
        self.atom_pool = spknn.base.Aggregate(axis=1, mean=False)

    def forward(self, inputs):
        atom_mask = inputs[Properties.atom_mask]
        neighbor_mask = inputs[Properties.neighbor_mask]
        distances = inputs["representation"]

        # Compute lennard jones potential
        power_6 = torch.where(
            neighbor_mask == 1,
            (self.r_equilibrium / distances) ** 6,
            torch.zeros_like(distances),
        )
        r_cut = self.cutoff(distances) * neighbor_mask

        yi = 0.5 * torch.sum((power_6 ** 2 - power_6) * r_cut, dim=2)[:, :, None]

        y = self.well_depth * self.atom_pool(yi, atom_mask)

        # collect results
        result = {self.property: y}

        if self.derivative is not None:
            sign = -1.0 if self.negative_dr else 1.0
            dy = grad(
                result[self.property],
                inputs[Properties.R],
                grad_outputs=torch.ones_like(result[self.property]),
                create_graph=self.create_graph,
                retain_graph=True,
            )[0]
            result[self.derivative] = sign * dy

        if self.stress is not None:
            cell = inputs[Properties.cell]
            # Compute derivative with respect to cell displacements
            stress = grad(
                result[self.property],
                inputs["displacement"],
                grad_outputs=torch.ones_like(result[self.property]),
                create_graph=self.create_graph,
                retain_graph=True,
            )[0]
            # Compute cell volume
            volume = torch.sum(
                cell[:, 0] * torch.cross(cell[:, 1], cell[:, 2]), dim=1, keepdim=True
            )[..., None]
            # Finalize stress tensor
            result[self.stress] = stress / volume

        return result


class CustomCutoff(nn.Module):
    def __init__(self, cutoff_radius, healing_length=0.3405):
        super(CustomCutoff, self).__init__()
        self.register_buffer("cutoff_radius", torch.Tensor([cutoff_radius]))
        self.register_buffer("healing_length", torch.Tensor([healing_length]))

    def forward(self, distances):
        r = (
            distances - (self.cutoff_radius - self.healing_length)
        ) / self.healing_length
        r_function = 1.0 + r ** 2 * (2.0 * r - 3.0)

        switch = torch.where(
            distances > self.cutoff_radius - self.healing_length,
            r_function,
            torch.ones_like(distances),
        )
        switch = torch.where(
            distances > self.cutoff_radius, torch.zeros_like(distances), switch
        )

        return switch


class LJCalculator(SchnetPackCalculator):
    def __init__(
        self,
        r_equilibrium,
        well_depth,
        required_properties,
        force_handle,
        position_conversion="Angstrom",
        force_conversion="eV / Angstrom",
        property_conversion={},
        stress_handle=None,
        stress_conversion="eV / Angstrom / Angstrom / Angstrom",
        detach=True,
        neighbor_list=spk.md.SimpleNeighborList,
        cutoff=-1.0,
        cutoff_shell=1.0,
        cutoff_lr=None,
        device="cpu",
    ):
        model = spk.atomistic.AtomisticModel(
            PairwiseRepresentation(),
            LJAtomistic(
                r_equilibrium=r_equilibrium,
                well_depth=well_depth,
                cutoff=cutoff,
                property=Properties.energy,
                negative_dr=True,
                derivative=force_handle,
                stress=stress_handle,
            ),
        ).to(device)

        super(LJCalculator, self).__init__(
            model,
            required_properties,
            force_handle,
            position_conversion=position_conversion,
            force_conversion=force_conversion,
            property_conversion=property_conversion,
            stress_handle=stress_handle,
            stress_conversion=stress_conversion,
            detach=detach,
            neighbor_list=neighbor_list,
            cutoff=cutoff,
            cutoff_shell=cutoff_shell,
            cutoff_lr=cutoff_lr,
        )



#!/usr/bin/env python
import os
import torch
import logging
import schnetpack as spk
from schnetpack.utils import (
    get_dataset,
    get_metrics,
    get_loaders,
    get_statistics,
    get_model,
    get_trainer,
    ScriptError,
    evaluate,
    setup_run,
    get_divide_by_atoms,
)
from schnetpack.utils.script_utils.settings import get_environment_provider
from schnetpack.utils.script_utils.parsing import build_parser

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


def main(args):
    # setup
    train_args = setup_run(args)

    device = torch.device("cuda" if args.cuda else "cpu")
    ep_device = torch.device(args.environment_provider_device)

    # get dataset
    environment_provider = get_environment_provider(train_args, device=ep_device)
    dataset = get_dataset(train_args, environment_provider=environment_provider)

    # get dataloaders
    split_path = os.path.join(args.modelpath, "split.npz")
    train_loader, val_loader, test_loader = get_loaders(
        args, dataset=dataset, split_path=split_path, logging=logging
    )

    # define metrics
    metrics = get_metrics(train_args)

    # train or evaluate
    if args.mode == "train":

        # get statistics
        atomref = dataset.get_atomref(args.property)
        mean, stddev = get_statistics(
            args=args,
            split_path=split_path,
            train_loader=train_loader,
            atomref=atomref,
            divide_by_atoms=get_divide_by_atoms(args),
            logging=logging,
        )

        # build model
        model = get_model(args, train_loader, mean, stddev, atomref, logging=logging)

        # build trainer
        logging.info("training...")
        trainer = get_trainer(args, model, train_loader, val_loader, metrics)

        # run training
        trainer.train(device, n_epochs=args.n_epochs)
        logging.info("...training done!")

    elif args.mode == "eval":

        # remove old evaluation files
        evaluation_fp = os.path.join(args.modelpath, "evaluation.txt")
        if os.path.exists(evaluation_fp):
            if args.overwrite:
                os.remove(evaluation_fp)
            else:
                raise ScriptError(
                    "The evaluation file does already exist at {}! Add overwrite flag"
                    " to remove.".format(evaluation_fp)
                )

        # load model
        logging.info("loading trained model...")
        model = spk.utils.load_model(
            os.path.join(args.modelpath, "best_model"), map_location=device
        )

        # run evaluation
        logging.info("evaluating...")
        if spk.utils.get_derivative(train_args) is None:
            with torch.no_grad():
                evaluate(
                    args,
                    model,
                    train_loader,
                    val_loader,
                    test_loader,
                    device,
                    metrics=metrics,
                )
        else:
            evaluate(
                args,
                model,
                train_loader,
                val_loader,
                test_loader,
                device,
                metrics=metrics,
            )
        logging.info("... evaluation done!")

    else:
        raise ScriptError("Unknown mode: {}".format(args.mode))


if __name__ == "__main__":
    parser = build_parser()
    args = parser.parse_args()

    if args.mode == "from_json":
        args = spk.utils.read_from_json(args.json_path)

    main(args)



import os
import time
import numpy as np
from schnetpack.train.hooks import Hook


__all__ = ["LoggingHook", "CSVHook", "TensorboardHook"]


class LoggingHook(Hook):
    """Base class for logging hooks.

    Args:
        log_path (str): path to directory in which log files will be stored.
        metrics (list): metrics to log; each metric has to be a subclass of spk.Metric.
        log_train_loss (bool, optional): enable logging of training loss.
        log_validation_loss (bool, optional): enable logging of validation loss.
        log_learning_rate (bool, optional): enable logging of current learning rate.

    """

    def __init__(
        self,
        log_path,
        metrics,
        log_train_loss=True,
        log_validation_loss=True,
        log_learning_rate=True,
    ):
        self.log_train_loss = log_train_loss
        self.log_validation_loss = log_validation_loss
        self.log_learning_rate = log_learning_rate
        self.log_path = log_path

        self._train_loss = 0
        self._counter = 0
        self.metrics = metrics

    def on_epoch_begin(self, trainer):
        """Log at the beginning of train epoch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.

        """
        # reset train_loss and counter
        if self.log_train_loss:
            self._train_loss = 0.0
            self._counter = 0
        else:
            self._train_loss = None

    def on_batch_end(self, trainer, train_batch, result, loss):
        if self.log_train_loss:
            n_samples = self._batch_size(result)
            self._train_loss += float(loss.data) * n_samples
            self._counter += n_samples

    def _batch_size(self, result):
        if type(result) is dict:
            n_samples = list(result.values())[0].size(0)
        elif type(result) in [list, tuple]:
            n_samples = result[0].size(0)
        else:
            n_samples = result.size(0)
        return n_samples

    def on_validation_begin(self, trainer):
        for metric in self.metrics:
            metric.reset()

    def on_validation_batch_end(self, trainer, val_batch, val_result):
        for metric in self.metrics:
            metric.add_batch(val_batch, val_result)


class CSVHook(LoggingHook):
    """Hook for logging training process to CSV files.

    Args:
        log_path (str): path to directory in which log files will be stored.
        metrics (list): metrics to log; each metric has to be a subclass of spk.Metric.
        log_train_loss (bool, optional): enable logging of training loss.
        log_validation_loss (bool, optional): enable logging of validation loss.
        log_learning_rate (bool, optional): enable logging of current learning rate.
        every_n_epochs (int, optional): epochs after which logging takes place.

    """

    def __init__(
        self,
        log_path,
        metrics,
        log_train_loss=True,
        log_validation_loss=True,
        log_learning_rate=True,
        every_n_epochs=1,
    ):
        log_path = os.path.join(log_path, "log.csv")
        super(CSVHook, self).__init__(
            log_path, metrics, log_train_loss, log_validation_loss, log_learning_rate
        )
        self._offset = 0
        self._restart = False
        self.every_n_epochs = every_n_epochs

    def on_train_begin(self, trainer):

        if os.path.exists(self.log_path):
            remove_file = False
            with open(self.log_path, "r") as f:
                # Ensure there is one entry apart from header
                lines = f.readlines()
                if len(lines) > 1:
                    self._offset = float(lines[-1].split(",")[0]) - time.time()
                    self._restart = True
                else:
                    remove_file = True

            # Empty up to header, remove to avoid adding header twice
            if remove_file:
                os.remove(self.log_path)
        else:
            self._offset = -time.time()
            # Create the log dir if it does not exists, since write cannot
            # create a full path
            log_dir = os.path.dirname(self.log_path)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)

        if not self._restart:
            log = ""
            log += "Time"

            if self.log_learning_rate:
                log += ",Learning rate"

            if self.log_train_loss:
                log += ",Train loss"

            if self.log_validation_loss:
                log += ",Validation loss"

            if len(self.metrics) > 0:
                log += ","

            for i, metric in enumerate(self.metrics):
                log += str(metric.name)
                if i < len(self.metrics) - 1:
                    log += ","

            with open(self.log_path, "a+") as f:
                f.write(log + os.linesep)

    def on_validation_end(self, trainer, val_loss):
        if trainer.epoch % self.every_n_epochs == 0:
            ctime = time.time() + self._offset
            log = str(ctime)

            if self.log_learning_rate:
                log += "," + str(trainer.optimizer.param_groups[0]["lr"])

            if self.log_train_loss:
                log += "," + str(self._train_loss / self._counter)

            if self.log_validation_loss:
                log += "," + str(val_loss)

            if len(self.metrics) > 0:
                log += ","

            for i, metric in enumerate(self.metrics):
                m = metric.aggregate()
                if hasattr(m, "__iter__"):
                    log += ",".join([str(j) for j in m])
                else:
                    log += str(m)
                if i < len(self.metrics) - 1:
                    log += ","

            with open(self.log_path, "a") as f:
                f.write(log + os.linesep)


class TensorboardHook(LoggingHook):
    """Hook for logging training process to tensorboard.

    Args:
        log_path (str): path to directory in which log files will be stored.
        metrics (list): metrics to log; each metric has to be a subclass of spk.Metric.
        log_train_loss (bool, optional): enable logging of training loss.
        log_validation_loss (bool, optional): enable logging of validation loss.
        log_learning_rate (bool, optional): enable logging of current learning rate.
        every_n_epochs (int, optional): epochs after which logging takes place.
        img_every_n_epochs (int, optional):
        log_histogram (bool, optional):

    """

    def __init__(
        self,
        log_path,
        metrics,
        log_train_loss=True,
        log_validation_loss=True,
        log_learning_rate=True,
        every_n_epochs=1,
        img_every_n_epochs=10,
        log_histogram=False,
    ):
        from tensorboardX import SummaryWriter

        super(TensorboardHook, self).__init__(
            log_path, metrics, log_train_loss, log_validation_loss, log_learning_rate
        )
        self.writer = SummaryWriter(self.log_path)
        self.every_n_epochs = every_n_epochs
        self.log_histogram = log_histogram
        self.img_every_n_epochs = img_every_n_epochs

    def on_epoch_end(self, trainer):
        if trainer.epoch % self.every_n_epochs == 0:
            if self.log_train_loss:
                self.writer.add_scalar(
                    "train/loss", self._train_loss / self._counter, trainer.epoch
                )
            if self.log_learning_rate:
                self.writer.add_scalar(
                    "train/learning_rate",
                    trainer.optimizer.param_groups[0]["lr"],
                    trainer.epoch,
                )

    def on_validation_end(self, trainer, val_loss):
        if trainer.epoch % self.every_n_epochs == 0:
            for metric in self.metrics:
                m = metric.aggregate()

                if np.isscalar(m):
                    self.writer.add_scalar(
                        "metrics/%s" % metric.name, float(m), trainer.epoch
                    )
                elif m.ndim == 2:
                    if trainer.epoch % self.img_every_n_epochs == 0:
                        import matplotlib.pyplot as plt

                        # tensorboardX only accepts images as numpy arrays.
                        # we therefore convert plots in numpy array
                        # see https://github.com/lanpa/tensorboard-pytorch/blob/master/examples/matplotlib_demo.py
                        fig = plt.figure()
                        plt.colorbar(plt.pcolor(m))
                        fig.canvas.draw()

                        np_image = np.fromstring(
                            fig.canvas.tostring_rgb(), dtype="uint8"
                        )
                        np_image = np_image.reshape(
                            fig.canvas.get_width_height()[::-1] + (3,)
                        )

                        plt.close(fig)

                        self.writer.add_image(
                            "metrics/%s" % metric.name, np_image, trainer.epoch
                        )

            if self.log_validation_loss:
                self.writer.add_scalar("train/val_loss", float(val_loss), trainer.step)

            if self.log_histogram:
                for name, param in trainer._model.named_parameters():
                    self.writer.add_histogram(
                        name, param.detach().cpu().numpy(), trainer.epoch
                    )

    def on_train_ends(self, trainer):
        self.writer.close()

    def on_train_failed(self, trainer):
        self.writer.close()



import torch
import schnetpack

__all__ = ["EnsembleCalculator"]


class EnsembleCalculator:
    def calculate(self, system):
        inputs = self._generate_input(system)

        results = []
        for model in self.models:
            prediction = model(inputs)
            results.append(prediction)

        # Compute statistics
        self.results = self._accumulate_results(results)
        self._update_system(system)

    def _accumulate_results(self, results):
        # Get the keys
        accumulated = {p: [] for p in results[0]}

        ensemble_results = {p: [] for p in results[0]}
        for p in accumulated:
            tmp = torch.stack([result[p].detach() for result in results])
            ensemble_results[p] = torch.mean(tmp, dim=0)
            ensemble_results["{:s}_var".format(p)] = torch.var(tmp, dim=0)

        return ensemble_results

    def _activate_stress(self, stress_handle):
        for model in self.models:
            schnetpack.utils.activate_stress_computation(model, stress=stress_handle)

    @staticmethod
    def _update_required_properties(required_properties):
        new_required = []
        for p in required_properties:
            prop_string = "{:s}_var".format(p)
            new_required += [p, prop_string]
            # Update properties
            # self.required_properties += [prop_string]
            # Update conversion
            # self.property_conversion[prop_string] = self.property_conversion[p]
        return new_required



"""
Instructions for interpreting molecular dynamics input files used in
``spk_md.py`` and initializing the associated classes.
"""
import schnetpack.md.calculators.schnet_calculator
import schnetpack.md.initial_conditions as initcond
import schnetpack.md.calculators as calculators
import schnetpack.md.integrators as integrators

import schnetpack.md.simulation_hooks.thermostats as thermostats
import schnetpack.md.simulation_hooks.barostats as barostats
import schnetpack.md.simulation_hooks.logging_hooks as logging_hooks
import schnetpack.md.simulation_hooks.sampling as sampling
from collections import OrderedDict


def model(x):
    """
    Dummy type for model class passed to CalculatorInit so basic conventions can be kept for other inputs

    Args:
        x (object): Torch model class
    Returns:
        object: The input model class
    """
    return x


def load_custom_calculator(calculator_module, class_name, **kwargs):
    """
    Function for loading custom calculators via the MD input file.

    This functionality can be accessed by specifying `custom` as the calculator type in the
    calculator block of the input file. It requires two inputs, the `calculator_module` and
    `class_name`. Both are used to specify how the custom calculator class (which should be based on the
    `schnetpack.md.calculators.MDCalculator`) is loaded. `calculator_module` can either be a path to the python
    file containing the calculator class or the name of a package/module if accessible via python. The `class_name`
    contains the name of the class which is loaded from the module. All additional inputs in the calculator block
    will be passed to the loaded custom calculator during initialization.

    Args:
        calculator_module (str): Can be either the path to the file containing the calculator module or the name
                                 of the module if it is installed.
        class_name (str): Name of the custom calculator class.
        **kwargs (optional): Additional arguments used during initialization. These can either be specific to the
                             custom calculator or be part of the `MDCalculator`.

    Returns:
        `schnetpack.md.calculators.MDCalculator`: Loaded calculator.
    """
    import os
    import importlib.util
    import importlib

    if os.path.exists(calculator_module):
        # First, check if a file is present and try to load the module
        module_spec = importlib.util.spec_from_file_location(
            "custom_calc", calculator_module
        )
        try:
            module_calc = importlib.util.module_from_spec(module_spec)
        except AttributeError:
            raise InitializerError(
                "Could not load module {:s}".format(calculator_module)
            )

        module_spec.loader.exec_module(module_calc)
    else:
        # Otherwise, try to import the module directly
        try:
            module_calc = importlib.import_module(calculator_module)
        except ModuleNotFoundError:
            raise InitializerError(
                "Could not find module {:s}".format(calculator_module)
            )

    # Get the calculator class from the module
    calculator = getattr(module_calc, class_name)
    # Initialize the calculator
    calculator = calculator(**kwargs)

    return calculator


class InitializerError(Exception):
    """
    Exception for SacredInit class.
    """

    pass


class Initializer:
    """
    Auxiliary class for setting up a molecular dynamics simulation with sacred in the run_dynamics.py script.
    Takes a dictionary of the form:
        ``init_dict = { 'type' : string identifying target class,
                      'name input1' : value,
                      'name_input2' : value,
                      ... }``

    It first checks the predefined allowed_options for the class associated with the short string given in type. Then
    it uses the associated input_type (another string) to check for the expected input pattern in another predefined
    dictionary required_inputs. The values provided in name_input... are then checked for consistency with the expected
    input pattern and converted to the appropriate types. Afterwards, the class is initialized using these inputs and
    stored in the 'initialized' variable.

    allowed_options (dict) should have entries of the form:
        string identifying target class : {
            reference to class,
            string indicating input type for lookup in required_inputs,
            optional kwargs, which can be used to set up specific default settings
        }

    required_inputs (dict) should have entries of the form:
        string indicating input type : OrderedDict({
            name : type
        })
    An ordered dict has to be used to preserve the input order passed to the class.

    Additional entries in the input_dict not associated with a conventional input are piped directly into the class as
    **kwargs and can be used to access additional options. This should be used with caution, as no consistency checks
    are performed in this case.

    Args:
        init_dict (dict): Dictionary containing input specifications generated in run_dynamics.py
    """

    kind = "type"
    allowed_options = {"test": ("class_name", "input_type", "optional_kwargs")}
    required_inputs = {"input_type": OrderedDict({"name_1": str, "name_2": int})}

    def __init__(self, init_dict):

        self.initialized = None

        if not init_dict:
            self.initialized = None
        else:
            kind = init_dict[self.kind]

            if kind not in self.allowed_options:
                raise InitializerError(
                    "Unrecognized type {:s} for {:s}. Options are:\n  {:s}".format(
                        kind, self.__class__.__name__, ", ".join(self.allowed_options)
                    )
                )

            target_class, input_type, optional_inputs = self.allowed_options[kind]

            required_inputs = self.required_inputs[input_type]

            inputs = []
            for required in required_inputs:
                # Check if required inputs are specified
                if required not in init_dict:
                    raise InitializerError(
                        "Expected input {:s} for {:s}.".format(required, kind)
                    )
                else:
                    # Try to convert to requested type
                    try:
                        inputs.append(required_inputs[required](init_dict[required]))
                    except ValueError:
                        raise InitializerError(
                            "Expected type {:s} for argument {:s}".format(
                                required_inputs[required].__name__, required
                            )
                        )

            # Additional inputs
            extra_kwargs = {}
            for key in init_dict:
                if key not in required_inputs:
                    if key != self.kind:
                        extra_kwargs[key] = init_dict[key]

            if optional_inputs is not None:
                extra_kwargs.update(optional_inputs)

            self.initialized = target_class(*inputs, **extra_kwargs)

    @classmethod
    def print_options(cls):
        """
        Print all options available for the initializer.
        """
        print("Available basic options:\n")
        for option in cls.allowed_options:
            input_type = cls.allowed_options[option][1]
            input_structure = cls.required_inputs[input_type]
            print("{:s}".format(option))
            print("-" * len(option))
            for k, v in input_structure.items():
                print(f"    {k} ({v.__name__})")
            print()


class ThermostatInit(Initializer):
    """
    Initialization instructions for thermostats. Optional instructions are used to define the massive and global
    Nose-Hoover chain thermostats. Two input cases exist, with all GLE based thermostats requiring an input file
    containing the thermostat matrices and all other types requiring specification of a time constant.
    """

    allowed_options = {
        "berendsen": (thermostats.BerendsenThermostat, "standard", None),
        "langevin": (thermostats.LangevinThermostat, "standard", None),
        "gle": (thermostats.GLEThermostat, "gle", None),
        "pile-l": (thermostats.PILELocalThermostat, "standard", None),
        "pile-g": (thermostats.PILEGlobalThermostat, "standard", None),
        "piglet": (thermostats.PIGLETThermostat, "gle", None),
        "nhc": (thermostats.NHCThermostat, "standard", None),
        "nhc-massive": (thermostats.NHCThermostat, "standard", {"massive": True}),
        "pi-nhc-l": (thermostats.NHCRingPolymerThermostat, "standard", None),
        "pi-nhc-g": (
            thermostats.NHCRingPolymerThermostat,
            "standard",
            {"local": False},
        ),
        "trpmd": (thermostats.TRPMDThermostat, "trpmd", None),
    }

    required_inputs = {
        "standard": OrderedDict({"temperature": float, "time_constant": float}),
        "gle": OrderedDict({"temperature": float, "gle_input": str}),
        "trpmd": OrderedDict({"temperature": float, "damping": float}),
    }


class BarostatInit(Initializer):
    """
    Initialization instructions for barostats. Optional instructions are used to define the
    Nose-Hoover chain thermostats.
    """

    allowed_options = {
        "nhc_iso": (barostats.NHCBarostatIsotropic, "standard", None),
        "nhc_aniso": (barostats.NHCBarostatAnisotropic, "standard", None),
        "pile": (barostats.RPMDBarostat, "standard", None),
    }

    required_inputs = {
        "standard": OrderedDict(
            {"target_pressure": float, "temperature": float, "time_constant": float}
        )
    }


class IntegratorInit(Initializer):
    """
    Initialization instructions for the available integrators.
    """

    allowed_options = {
        "verlet": (integrators.VelocityVerlet, "verlet", None),
        "ring_polymer": (integrators.RingPolymer, "ring_polymer", None),
        "npt_verlet": (integrators.NPTVelocityVerlet, "npt_verlet", None),
        "npt_ring_polymer": (integrators.NPTRingPolymer, "npt_ring_polymer", None),
    }

    required_inputs = {
        "verlet": OrderedDict({"time_step": float}),
        "ring_polymer": OrderedDict(
            {"n_beads": int, "time_step": float, "temperature": float}
        ),
        "npt_verlet": OrderedDict({"time_step": float}),
        "npt_ring_polymer": OrderedDict(
            {"n_beads": int, "time_step": float, "temperature": float}
        ),
    }


class InitialConditionsInit(Initializer):
    """
    Initialization instructions for the available thermostats.
    """

    allowed_options = {"maxwell-boltzmann": (initcond.MaxwellBoltzmannInit, "mb", None)}
    required_inputs = {
        "mb": OrderedDict(
            {"temperature": float, "remove_translation": bool, "remove_rotation": bool}
        )
    }


class CalculatorInit(Initializer):
    """
    Initialization instructions for the available calculators.
    """

    allowed_options = {
        "schnet": (
            schnetpack.md.calculators.schnet_calculator.SchnetPackCalculator,
            "schnet",
            None,
        ),
        "schnet_ensemble": (
            schnetpack.md.calculators.schnet_calculator.EnsembleSchnetPackCalculator,
            "schnet_ensemble",
            None,
        ),
        "orca": (calculators.OrcaCalculator, "orca", None),
        "sgdml": (calculators.SGDMLCalculator, "sgdml", None),
        "custom": (load_custom_calculator, "custom", None),
        "custom_ensemble": (load_custom_calculator, "custom_ensemble", None),
    }
    required_inputs = {
        "schnet": OrderedDict(
            {"model": model, "required_properties": list, "force_handle": str}
        ),
        "schnet_ensemble": OrderedDict(
            {"models": list, "required_properties": list, "force_handle": str}
        ),
        "orca": OrderedDict(
            {
                "required_properties": list,
                "force_handle": str,
                "compdir": str,
                "qm_executable": str,
                "orca_template": str,
            }
        ),
        "sgdml": OrderedDict({"model": model}),
        "custom": OrderedDict({"calculator_module": str, "class_name": str}),
        "custom_ensemble": OrderedDict({"calculator_module": str, "class_name": str}),
    }


class BiasPotentialInit(Initializer):
    """
    Set up the available bias potentials
    """

    allowed_options = {
        "accelerated_md": (sampling.AcceleratedMD, "accelerated_md", None),
        "metadyn": (sampling.MetaDyn, "metadyn", None),
    }
    required_inputs = {
        "accelerated_md": OrderedDict(
            {"energy_threshold": float, "acceleration_factor": float}
        ),
        "metadyn": OrderedDict({"collective_variables": list}),
    }


class ColVars:
    available = {"bond": sampling.BondColvar}


class AdaptiveSamplingInit(Initializer):
    """
    Set up directives for adaptive sampling
    """

    allowed_options = {"ensemble": (sampling.AdaptiveSampling, "ensemble", None)}
    required_inputs = {
        "ensemble": OrderedDict({"thresholds": dict, "n_samples": int, "dataset": str})
    }


class LoggerStreams:
    """
    Auxiliary class to provide short strings identifying the different available data streams for the FileLogger hook.
    """

    available = {
        "molecules": logging_hooks.MoleculeStream(),
        "properties": logging_hooks.PropertyStream(),
        "dynamic": logging_hooks.SimulationStream(),
    }


def get_data_streams(stream_list):
    """
    Auxiliary function to translate a list of identifier strings specified in LoggerStream.available into the
    corresponding DataStream classes to be used in the file logger.

    Args:
        stream_list: List of strings indicating the requested DataStreams for the file logger.

    Returns:
        list(object): List of DataStream objects.
    """
    streams = []
    for stream in stream_list:
        if stream not in LoggerStreams.available:
            raise ValueError("Stream {:s} not available for file logger".format(stream))
        streams.append(LoggerStreams.available[stream])

    return streams



import os

import schnetpack as spk
import torch
from torch.optim import Adam

__all__ = ["get_trainer", "simple_loss_fn", "tradeoff_loss_fn", "get_metrics"]


def get_trainer(args, model, train_loader, val_loader, metrics):
    # setup optimizer
    # filter for trainable parameters (https://github.com/pytorch/pytorch/issues/679)
    trainable_params = filter(lambda p: p.requires_grad, model.parameters())
    optimizer = Adam(trainable_params, lr=args.lr)

    # setup hook and logging
    hooks = [spk.train.MaxEpochHook(args.max_epochs)]
    if args.max_steps:
        hooks.append(spk.train.MaxStepHook(max_steps=args.max_steps))

    schedule = spk.train.ReduceLROnPlateauHook(
        optimizer=optimizer,
        patience=args.lr_patience,
        factor=args.lr_decay,
        min_lr=args.lr_min,
        window_length=1,
        stop_after_min=True,
    )
    hooks.append(schedule)

    if args.logger == "csv":
        logger = spk.train.CSVHook(
            os.path.join(args.modelpath, "log"),
            metrics,
            every_n_epochs=args.log_every_n_epochs,
        )
        hooks.append(logger)
    elif args.logger == "tensorboard":
        logger = spk.train.TensorboardHook(
            os.path.join(args.modelpath, "log"),
            metrics,
            every_n_epochs=args.log_every_n_epochs,
        )
        hooks.append(logger)

    # setup loss function
    loss_fn = get_loss_fn(args)

    # setup trainer
    trainer = spk.train.Trainer(
        args.modelpath,
        model,
        loss_fn,
        optimizer,
        train_loader,
        val_loader,
        checkpoint_interval=args.checkpoint_interval,
        keep_n_checkpoints=args.keep_n_checkpoints,
        hooks=hooks,
    )
    return trainer


def get_loss_fn(args):
    derivative = spk.utils.get_derivative(args)
    contributions = spk.utils.get_contributions(args)
    stress = spk.utils.get_stress(args)

    # simple loss function for training on property only
    if derivative is None and contributions is None and stress is None:
        return simple_loss_fn(args)

    # loss function with tradeoff weights
    if type(args.rho) == float:
        rho = dict(property=args.rho, derivative=1 - args.rho)
    else:
        rho = dict()
        rho["property"] = (
            1.0 if "property" not in args.rho.keys() else args.rho["property"]
        )
        if derivative is not None:
            rho["derivative"] = (
                1.0 if "derivative" not in args.rho.keys() else args.rho["derivative"]
            )
        if contributions is not None:
            rho["contributions"] = (
                1.0
                if "contributions" not in args.rho.keys()
                else args.rho["contributions"]
            )
        if stress is not None:
            rho["stress"] = (
                1.0 if "stress" not in args.rho.keys() else args.rho["stress"]
            )
        # type cast of rho values
        for key in rho.keys():
            rho[key] = float(rho[key])
        # norm rho values
        norm = sum(rho.values())
        for key in rho.keys():
            rho[key] = rho[key] / norm
    property_names = dict(
        property=args.property,
        derivative=derivative,
        contributions=contributions,
        stress=stress,
    )
    return tradeoff_loss_fn(rho, property_names)


def simple_loss_fn(args):
    def loss(batch, result):
        diff = batch[args.property] - result[args.property]
        diff = diff ** 2
        err_sq = torch.mean(diff)
        return err_sq

    return loss


def tradeoff_loss_fn(rho, property_names):
    def loss(batch, result):
        err = 0.0
        for prop, tradeoff_weight in rho.items():
            # TODO: contributions should not be here
            diff = batch[property_names[prop]] - result[property_names[prop]]
            diff = diff ** 2
            err += tradeoff_weight * torch.mean(diff)

        return err

    return loss


def get_metrics(args):
    # setup property metrics
    metrics = [
        spk.train.metrics.MeanAbsoluteError(args.property, args.property),
        spk.train.metrics.RootMeanSquaredError(args.property, args.property),
    ]

    # add metrics for derivative
    derivative = spk.utils.get_derivative(args)
    if derivative is not None:
        metrics += [
            spk.train.metrics.MeanAbsoluteError(
                derivative, derivative, element_wise=True
            ),
            spk.train.metrics.RootMeanSquaredError(
                derivative, derivative, element_wise=True
            ),
        ]

    # Add stress metric
    stress = spk.utils.get_stress(args)
    if stress is not None:
        metrics += [
            spk.train.metrics.MeanAbsoluteError(stress, stress, element_wise=True),
            spk.train.metrics.RootMeanSquaredError(stress, stress, element_wise=True),
        ]

    return metrics



"""
Module for setting up the initial conditions of the molecules in :obj:`schnetpack.md.System`.
This entails sampling the momenta from random distributions corresponding to certain temperatures.
"""
import torch

__all__ = ["Initializer", "MaxwellBoltzmannInit"]


class Initializer:
    """
    Basic initializer class template. Initializes the systems momenta to correspond to a certain temperature.

    Args:
        temperature (float):  Target initialization temperature in Kelvin.
    """

    def __init__(self, temperature):
        self.temperature = temperature

    def initialize_system(self, system, mask=None):
        """
        Initialize the system according to the instructions given in _setup_momenta.

        Args:
            system (object): System class containing all molecules and their replicas.
            mask (array): Mask where 1 indicates that velocities of system will be changed.
        """
        self._setup_momenta(system, mask=mask)

    def _setup_momenta(self, system, mask=None):
        """
        Main routine for initializing system momenta based on the molecules defined in system and the provided
        temperature. To be implemented.

        Args:
            system (object): System class containing all molecules and their replicas.
            mask (array): Mask where 1 indicates that velocities of system will be changed.
        """
        raise NotImplementedError


class MaxwellBoltzmannInit(Initializer):
    """
    Initializes the system momenta according to a Maxwell--Boltzmann distribution at the given temperature.

    Args:
        temperature (float): Target temperature in Kelvin.
        remove_translation (bool): Remove the translational components of the momenta after initialization. Will stop
                                   molecular drift for NVE simulations and NVT simulations with deterministic
                                   thermostats (default=False).
        remove_rotation (bool): Remove the rotational components of the momenta after initialization. Will reduce
                                molecular rotation for NVE simulations and NVT simulations with deterministic
                                thermostats (default=False).
    """

    def __init__(self, temperature, remove_translation=False, remove_rotation=False):
        super(MaxwellBoltzmannInit, self).__init__(temperature)
        self.remove_translation = remove_translation
        self.remove_rotation = remove_rotation

    def _setup_momenta(self, system, mask=None):
        """
        Initialize the momenta, by drawing from a random normal distribution and rescaling the momenta to the desired
        temperature afterwards. In addition, the system is centered at its center of mass.

        Args:
            system (object): System class containing all molecules and their replicas.
        """
        # Move center of mass to origin
        system.remove_com()

        # Set initial system momenta and apply atom masks
        momenta = (
            torch.randn(system.momenta.shape, device=system.device)
            * system.masses
            * system.atom_masks
        )

        if mask is None:
            system.momenta = momenta
        else:
            system.momenta[mask == 1] = momenta[mask == 1]

        # Remove translational motion if requested
        if self.remove_translation:
            system.remove_com_translation()

        # Remove rotational motion if requested
        if self.remove_rotation:
            system.remove_com_rotation()

        # Scale velocities to desired temperature
        scaling = torch.sqrt(self.temperature / system.temperature)
        if mask is None:
            system.momenta *= scaling[:, :, None, None]
        else:
            system.momenta[mask == 1, ...] *= scaling[mask == 1][:, None, None]



"""
Module for setting up a :obj:`schnetpack.md.simulator.Simulator` from a config dictionary,
which can then be used for simulation.
"""
import logging
import os
import torch
from shutil import rmtree
from ase import io

try:
    import oyaml as yaml
except ImportError:
    import yaml

from schnetpack.md import Simulator, System
from schnetpack.md.parsers.md_options import *
from schnetpack.md.simulation_hooks import *
from schnetpack.md.neighbor_lists import *


class MDSimulation:
    """
    Set up a :obj:`schnetpack.md.Simulator` for performing molecular dynamics simulations
    based on a configuration dictionary. The dictionary is parsed using individual
    :obj:`SetupBlock` instances, which also update the dictionary if default settings are changed.

    Args:
        config (dict): Configuration dictionary either read from an input
                       file in yaml format or generated internally.
    """

    def __init__(self, config):
        self.config = config

        # Setup containers
        self.device = None
        self.simulation_dir = None
        self.integrator = None
        self.system = None
        self.calculator = None
        self.restart = False
        self.load_system_state = False
        self.n_steps = None
        self.hooks = []

        # Initialize device
        SetupDevice(self)

        # Setup seed
        SetupSeed(self)

        # Setup directories
        SetupDirectories(self)

        # Setup the system, including initial conditions
        SetupSystem(self)

        # Get the calculator
        SetupCalculator(self)

        # Get the integrator
        SetupDynamics(self)

        # Get bias potentials if applicable
        SetupBiasPotential(self)

        # Set up adaptive sampling
        SetupAdaptiveSampling(self)

        # Setup Logging
        SetupLogging(self)

        # Get simulator
        self.simulator = self._build_simulator()

    def _build_simulator(self):
        """
        Build the final :obj:`schnetpack.md.Simulator` based on all collected
        subroutines.

        Returns:
            schnetpack.md.Simulator: The simulator for performing the MD.
        """
        simulator = Simulator(self.system, self.integrator, self.calculator, self.hooks)

        # If requested, read restart data
        if self.restart and (self.restart is not None):
            state_dict = torch.load(self.restart, map_location=self.device)
            simulator.restart_simulation(state_dict, soft=False)
            logging.info(f"Restarting simulation from {self.restart}...")
        elif self.load_system_state:
            state_dict = torch.load(self.load_system_state, map_location=self.device)
            simulator.load_system_state(state_dict)
            logging.info(f"Loaded system state from {self.load_system_state}...")

        return simulator

    def run(self):
        """
        Run the simulation using the internal simulator.
        """
        self.simulator.simulate(self.n_steps)

    def save_config(self):
        """
        Save the collected and updated config directory to a yaml file.
        """
        yamlpath = os.path.join(self.simulation_dir, "config.yaml")
        with open(yamlpath, "w") as yf:
            yaml.dump(self.config, yf, default_flow_style=False)


class SetupBlock:
    """
    Base class for defining default entries in the config dictionary as well as
    parsing new instructions. Based on these instructions, the ``_setup`` function is
    used to construct the target simulation module or carry out the requested tasks.
    At the same time, existing config entries are updated.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {}
    target_block = None

    def __init__(self, md_initializer):
        self.target_config_block = self._get_target_block(md_initializer)
        self._load_defaults()
        self._setup(md_initializer)

    def _load_defaults(self):
        """
        Check for default settings in ``self.default_options`` and update config dictionary.
        """
        for option in self.default_options:
            if option not in self.target_config_block:
                self.target_config_block[option] = self.default_options[option]

            elif type(self.default_options[option]) == dict:
                for sub_option in self.default_options[option]:
                    if sub_option not in self.target_config_block[option]:
                        self.target_config_block[option][
                            sub_option
                        ] = self.default_options[option][sub_option]

    def _get_target_block(self, md_initializer):
        """
        Determine the target block the initializer is acting on.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        if self.target_block is not None:
            if self.target_block in md_initializer.config:
                return md_initializer.config[self.target_block]
            else:
                return None
        else:
            return md_initializer.config

    def _setup(self, md_initializer):
        """
        Placeholder function. This is used to set up the individual simulation
        modules and needs to be adapted accordingly.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        raise NotImplementedError


class SetupDirectories(SetupBlock):
    """
    Routine for setting up simulation directories.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {"simulation_dir": "simulation_dir", "overwrite": True}

    def _setup(self, md_initializer):
        """
        Main setup routine, check if directory exists and perform appropriate actions.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        simulation_dir = self.target_config_block["simulation_dir"]
        overwrite = self.target_config_block["overwrite"]

        logging.info("Create model directory")

        if simulation_dir is None:
            raise ValueError("Config `simulation_dir` has to be set!")

        if os.path.exists(simulation_dir) and not overwrite:
            logging.warning(
                "Simulation directory {:s} already exists (set overwrite flag?)".format(
                    simulation_dir
                )
            )

        if os.path.exists(simulation_dir) and overwrite:
            rmtree(simulation_dir)

        if not os.path.exists(simulation_dir):
            os.makedirs(simulation_dir)

        md_initializer.simulation_dir = simulation_dir


class SetupSeed(SetupBlock):
    """
    Initialize the random seed from the config file.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {"seed": None}

    def _setup(self, md_initializer):
        """
        Initialize the random seed from the config file.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        seed = self.target_config_block["seed"]
        from schnetpack.utils import set_random_seed

        set_random_seed(seed)


class SetupDevice(SetupBlock):
    """
    Determine the computation device.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {"device": "cpu"}

    def _setup(self, md_initializer):
        """
        Determine the computation device from the config.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        device = torch.device(md_initializer.config["device"])
        md_initializer.device = device


class SetupSystem(SetupBlock):
    """
    Parse the system block in the input file and setup the
    atomistic system in :obj:`schnetpack.md.System`.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {
        "n_replicas": 1,
        "molecule_file": "dummy.xyz",
        "initializer": {
            InitialConditionsInit.kind: "maxwell-boltzmann",
            "temperature": 300,
            "remove_translation": True,
            "remove_rotation": True,
        },
        "load_system_state": False,
    }
    target_block = "system"

    def _setup(self, md_initializer):
        """
        Main routine. Loads the molecules and initializes the system.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        molecule_file = self.target_config_block["molecule_file"]
        n_replicas = self.target_config_block["n_replicas"]

        # Read in the molecular structures
        ase_molecules = io.read(molecule_file, index=":")

        # Set up the system
        system = System(n_replicas, device=md_initializer.device)
        system.load_molecules(ase_molecules)

        # Apply initial conditions if requested
        if "initializer" in self.target_config_block:
            initializer = self.target_config_block["initializer"]
            initconds = InitialConditionsInit(initializer)

            if initconds.initialized is not None:
                initconds.initialized.initialize_system(system)

        md_initializer.load_system_state = self.target_config_block["load_system_state"]
        md_initializer.system = system


class SetupCalculator(SetupBlock):
    """
    Parse the calculator block and initialize the calculator for the molecular dynamics simulations.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {
        CalculatorInit.kind: "schnet",
        "required_properties": ["energy", "forces"],
        "force_handle": "forces",
        # "position_conversion": "Angstrom",
        # "force_conversion": "kcal/mol/Angstrom",
        "property_conversion": {},
        # "neighbor_list": "simple",
        # "cutoff": -1.0,
        # "cutoff_shell": 1.0,
    }
    target_block = "calculator"
    basic_models = ["schnet", "custom"]
    ensemble_models = ["schnet_ensemble", "custom_ensemble"]
    neighbor_list = {
        "simple": SimpleNeighborList,
        "ase": ASENeighborList,
        "torch": TorchNeighborList,
    }

    def _setup(self, md_initializer):
        """
        Main routine for loading the model, preparing it for the calculator and setting
        up the main :obj:`schnetpack.md.calculator`.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        calculator = self.target_config_block
        calculator_dict = {}

        # Load model, else get options
        for key in calculator:
            # TODO Figure out conventions for ensemble
            if key == "model_file" or key == "model_files":
                if calculator[CalculatorInit.kind] in self.basic_models:
                    model = self._load_model_schnetpack(
                        calculator["model_file"], md_initializer.device
                    ).to(md_initializer.device)
                    calculator_dict["model"] = model
                elif calculator[CalculatorInit.kind] in self.ensemble_models:
                    models = [
                        self._load_model_schnetpack(model, md_initializer.device).to(
                            md_initializer.device
                        )
                        for model in calculator["model_files"]
                    ]
                    calculator_dict["models"] = models
                elif calculator[CalculatorInit.kind] == "sgdml":
                    model = self._load_model_sgdml(calculator["model_file"]).to(
                        md_initializer.device
                    )
                    calculator_dict["model"] = model
                else:
                    raise ValueError(
                        f"Unrecognized ML calculator {calculator[CalculatorInit.kind]}"
                    )
            elif key == "neighbor_list":
                # Check for neighbor list
                calculator_dict[key] = self.neighbor_list[calculator[key]]
            else:
                calculator_dict[key] = calculator[key]

        calculator = CalculatorInit(calculator_dict).initialized

        md_initializer.calculator = calculator

    @staticmethod
    def _load_model_schnetpack(model_path, device):
        """
        Load the model from a model file and move it to the desired device.

        Args:
            model_path (str): Path to the stored model.
            device (torch.device): Device the computations should be performed on.

        Returns:
            schnetpack.AtomisticModel: SchNetPack model.

        """
        # If model is a directory, search for best_model file
        if os.path.isdir(model_path):
            model_path = os.path.join(model_path, "best_model")

        # Load model
        model = schnetpack.utils.load_model(model_path, map_location=device)
        logging.info("Loaded model from {:s}".format(model_path))

        return model

    @staticmethod
    def _load_model_sgdml(model_path):
        """
        Load an sGDML model from disk.

        Args:
            model_path (str): Path to the stored sGDML model parameters.

        Returns:
            sgdml.torchtools.GDMLTorchPredict: sGDML model.
        """
        import numpy as np

        try:
            from sgdml.torchtools import GDMLTorchPredict
        except:
            raise ImportError(
                "Could not load sGDML. Please make sure the package is installed."
            )

        try:
            parameters = np.load(model_path, allow_pickle=True)
        except:
            raise ValueError("Could not read sGDML model from {:s}".format(model_path))

        model = GDMLTorchPredict(parameters)
        logging.info("Loaded sGDML model from {:s}".format(model_path))
        return model


class SetupDynamics(SetupBlock):
    """
    Parse the `dynamics` block in the input file and prepare the :obj:`schnetpack.md.Integrator`.
    Also parses thermostats and whether rotational and translational motion should be removed
    during simulation.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {
        "integrator": {IntegratorInit.kind: "verlet", "time_step": 0.5},
        "n_steps": 10000,
        "restart": False,
    }
    target_block = "dynamics"

    def _setup(self, md_initializer):
        """
        Initialize the integrator, as well as all thermostat hooks.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """
        # Get the integrator config
        integrator_config = self.target_config_block["integrator"]

        # Set the device for integrator (used in normal mode transformation)
        integrator_config["device"] = md_initializer.device

        if integrator_config[IntegratorInit.kind] == "ring_polymer":
            integrator_config["n_beads"] = md_initializer.system.n_replicas

        # Add a thermostat if requested (certain baristats have thermostats built in,
        # in this case all options here are ignored)
        if "thermostat" in self.target_config_block:
            thermostat_config = self.target_config_block["thermostat"]
            thermostat = ThermostatInit(thermostat_config).initialized
        else:
            thermostat = None

        if "barostat" in self.target_config_block:
            barostat_config = self.target_config_block["barostat"]

            # Issue message in case a barostat with included thermostat is used
            # -> NHC type barostats
            if barostat_config[BarostatInit.kind] in ["nhc_iso", "nhc_aniso"]:
                logging.info(
                    "The chosen barostat {:s} already includes a thermostat."
                    "All thermostat options will be ignored".format(BarostatInit.kind)
                )
                thermostat = None

            # Automatically set barostat temperature to thermostat temperature if feasible
            if thermostat is not None:
                logging.info(
                    f"Thermostat detected, setting barostat temperature to {thermostat.temperature_bath}K"
                )
                barostat_config["temperature"] = thermostat.temperature_bath

            barostat = BarostatInit(barostat_config).initialized

            # If a barostat of the NHC and BLA kind is used, switch to the corresponding NPT integrator
            if barostat_config is not None:
                if integrator_config[IntegratorInit.kind] == "ring_polymer":
                    integrator_config[IntegratorInit.kind] = "npt_ring_polymer"
                    integrator_config["barostat"] = barostat
                if integrator_config[IntegratorInit.kind] == "verlet":
                    integrator_config[IntegratorInit.kind] = "npt_verlet"
                    integrator_config["barostat"] = barostat
        else:
            barostat = None

        # Finally add barostat and thermostats to hooks
        if barostat is not None:
            md_initializer.hooks.append(barostat)
        if thermostat is not None:
            md_initializer.hooks.append(thermostat)

        # Iniotialize the integrator based on the above modifications
        md_initializer.integrator = IntegratorInit(integrator_config).initialized

        # Remove the motion of the center of motion
        if "remove_com_motion" in self.target_config_block:
            remove_com_config = self.target_config_block["remove_com_motion"]

            if (
                "every_n_steps" not in remove_com_config
                or "remove_rotation" not in remove_com_config
            ):
                raise InitializerError("Missing options in remove_com_motion")
            else:
                md_initializer.hooks += [
                    RemoveCOMMotion(
                        every_n_steps=remove_com_config["every_n_steps"],
                        remove_rotation=remove_com_config["remove_rotation"],
                    )
                ]

        md_initializer.n_steps = self.target_config_block["n_steps"]
        md_initializer.restart = self.target_config_block["restart"]


class SetupBiasPotential(SetupBlock):
    """
    Routine for building a bias potential hook if requested.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {}
    target_block = "bias_potential"

    def _setup(self, md_initializer):
        """
        Check if a bias potential is present and parse the options depending on the type
        of potential. The appropriate hook is then added to those already present.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """

        if "bias_potential" in md_initializer.config:
            bias_potential = self.target_config_block

            # For metadynamics
            if bias_potential[BiasPotentialInit.kind] == "metadyn":
                dummy_potential = {}
                colvars = []
                # Parse the provided collective variable options
                for k, v in bias_potential.items():
                    if k == "colvars":
                        for cv in bias_potential[k]:
                            cv = cv.split()
                            cv_type = cv[0].lower()
                            cv_inputs = [int(x) for x in cv[1:3]]
                            cv_width = float(cv[3])
                            colvars.append(
                                ColVars.available[cv_type](*cv_inputs, cv_width)
                            )
                        dummy_potential["collective_variables"] = colvars
                    else:
                        dummy_potential[k] = v

                md_initializer.hooks += [BiasPotentialInit(dummy_potential).initialized]
            # For accelerated molecular dynamics
            else:
                md_initializer.hooks += [BiasPotentialInit(bias_potential).initialized]


class SetupAdaptiveSampling(SetupBlock):
    default_options = {}
    target_block = "adaptive_sampling"

    def _setup(self, md_initializer):
        if self.target_block in md_initializer.config:
            adaptive_sampling = self.target_config_block

            if adaptive_sampling[AdaptiveSamplingInit.kind] == "ensemble":
                with open(adaptive_sampling["thresholds"], "r") as tf:
                    thresholds = yaml.load(tf)
                adaptive_sampling["thresholds"] = thresholds

            md_initializer.hooks += [
                AdaptiveSamplingInit(adaptive_sampling).initialized
            ]


class SetupLogging(SetupBlock):
    """
    Construct and update the logging hooks.

    Args:
        md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
    """

    default_options = {
        "file_logger": {
            "buffer_size": 100,
            "streams": ["molecules", "properties"],
            "every_n_steps": 1,
        },
        "temperature_logger": 100,
        "write_checkpoints": 1000,
    }
    target_block = "logging"

    def _setup(self, md_initializer):
        """
        Parse the config and create all requested logging hooks.

        Args:
            md_initializer (schnetpack.md.parser.MDSimulation): Parent MDSimulation class.
        """

        if "file_logger" in self.target_config_block:
            logging_file = os.path.join(
                md_initializer.simulation_dir, "simulation.hdf5"
            )
            file_logging_config = self.target_config_block["file_logger"]

            data_streams = get_data_streams(file_logging_config["streams"])

            md_initializer.hooks += [
                FileLogger(
                    logging_file,
                    file_logging_config["buffer_size"],
                    data_streams=data_streams,
                    every_n_steps=file_logging_config["every_n_steps"],
                )
            ]

        tensorboard_logdir = os.path.join(md_initializer.simulation_dir, "logging")

        if "temperature_logger" in self.target_config_block:
            # temperature_dir = os.path.join(md_initializer.simulation_dir, "temperature")
            temperature_dir = tensorboard_logdir
            md_initializer.hooks += [
                TemperatureLogger(
                    temperature_dir,
                    every_n_steps=self.target_config_block["temperature_logger"],
                )
            ]

        if "pressure_logger" in self.target_config_block:
            pressure_dir = tensorboard_logdir
            md_initializer.hooks += [
                PressureLogger(
                    pressure_dir,
                    every_n_steps=self.target_config_block["pressure_logger"],
                )
            ]

        if "write_checkpoints" in self.target_config_block:
            chk_file = os.path.join(md_initializer.simulation_dir, "checkpoint.chk")
            md_initializer.hooks += [
                Checkpoint(
                    chk_file,
                    every_n_steps=self.target_config_block["write_checkpoints"],
                )
            ]



import torch
from torch import nn

from schnetpack.nn import Dense
from schnetpack.nn.base import Aggregate


__all__ = ["CFConv"]


class CFConv(nn.Module):
    r"""Continuous-filter convolution block used in SchNet module.

    Args:
        n_in (int): number of input (i.e. atomic embedding) dimensions.
        n_filters (int): number of filter dimensions.
        n_out (int): number of output dimensions.
        filter_network (nn.Module): filter block.
        cutoff_network (nn.Module, optional): if None, no cut off function is used.
        activation (callable, optional): if None, no activation function is used.
        normalize_filter (bool, optional): If True, normalize filter to the number
            of neighbors when aggregating.
        axis (int, optional): axis over which convolution should be applied.

    """

    def __init__(
        self,
        n_in,
        n_filters,
        n_out,
        filter_network,
        cutoff_network=None,
        activation=None,
        normalize_filter=False,
        axis=2,
    ):
        super(CFConv, self).__init__()
        self.in2f = Dense(n_in, n_filters, bias=False, activation=None)
        self.f2out = Dense(n_filters, n_out, bias=True, activation=activation)
        self.filter_network = filter_network
        self.cutoff_network = cutoff_network
        self.agg = Aggregate(axis=axis, mean=normalize_filter)

    def forward(self, x, r_ij, neighbors, pairwise_mask, f_ij=None):
        """Compute convolution block.

        Args:
            x (torch.Tensor): input representation/embedding of atomic environments
                with (N_b, N_a, n_in) shape.
            r_ij (torch.Tensor): interatomic distances of (N_b, N_a, N_nbh) shape.
            neighbors (torch.Tensor): indices of neighbors of (N_b, N_a, N_nbh) shape.
            pairwise_mask (torch.Tensor): mask to filter out non-existing neighbors
                introduced via padding.
            f_ij (torch.Tensor, optional): expanded interatomic distances in a basis.
                If None, r_ij.unsqueeze(-1) is used.

        Returns:
            torch.Tensor: block output with (N_b, N_a, n_out) shape.

        """
        if f_ij is None:
            f_ij = r_ij.unsqueeze(-1)

        # pass expanded interactomic distances through filter block
        W = self.filter_network(f_ij)
        # apply cutoff
        if self.cutoff_network is not None:
            C = self.cutoff_network(r_ij)
            W = W * C.unsqueeze(-1)

        # pass initial embeddings through Dense layer
        y = self.in2f(x)
        # reshape y for element-wise multiplication by W
        nbh_size = neighbors.size()
        nbh = neighbors.view(-1, nbh_size[1] * nbh_size[2], 1)
        nbh = nbh.expand(-1, -1, y.size(2))
        y = torch.gather(y, 1, nbh)
        y = y.view(nbh_size[0], nbh_size[1], nbh_size[2], -1)

        # element-wise multiplication, aggregating and Dense layer
        y = y * W
        y = self.agg(y, pairwise_mask)
        y = self.f2out(y)
        return y



"""
This module contains all functionality for performing various molecular dynamics simulations
using SchNetPack.
"""
from .system import *
from .initial_conditions import *
from .integrators import *
from .neighbor_lists import *
from .simulator import *
from .utils import *



import torch
import numpy as np
from torch.optim.lr_scheduler import ReduceLROnPlateau, StepLR
from schnetpack.train.hooks import Hook


class EarlyStoppingHook(Hook):
    r"""Hook to stop training if validation loss fails to improve.

    Args:
        patience (int): number of epochs which can pass without improvement
            of validation loss before training ends.
        threshold_ratio (float, optional): counter increases if
            curr_val_loss > (1-threshold_ratio) * best_loss

    """

    def __init__(self, patience, threshold_ratio=0.0001):
        self.best_loss = float("Inf")
        self.counter = 0
        self.threshold_ratio = threshold_ratio
        self.patience = patience

    @property
    def state_dict(self):
        return {"counter": self.counter, "best_loss": self.best_loss}

    @state_dict.setter
    def state_dict(self, state_dict):
        self.counter = state_dict["counter"]
        self.best_loss = state_dict["best_loss"]

    def on_validation_end(self, trainer, val_loss):
        if val_loss > (1 - self.threshold_ratio) * self.best_loss:
            self.counter += 1
        else:
            self.best_loss = val_loss
            self.counter = 0

        if self.counter > self.patience:
            trainer._stop = True


class WarmRestartHook(Hook):
    def __init__(
        self, T0=10, Tmult=2, each_step=False, lr_min=1e-6, lr_factor=1.0, patience=1
    ):
        self.scheduler = None
        self.each_step = each_step
        self.T0 = T0
        self.Tmult = Tmult
        self.Tmax = T0
        self.lr_min = lr_min
        self.lr_factor = lr_factor
        self.patience = patience
        self.waiting = 0

        self.best_previous = float("Inf")
        self.best_current = float("Inf")

    def on_train_begin(self, trainer):
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            trainer.optimizer, self.Tmax, self.lr_min
        )
        self.init_opt_state = trainer.optimizer.state_dict()

    def on_batch_begin(self, trainer, train_batch):
        """Log at the beginning of train batch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.
            train_batch (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        """
        if self.each_step:
            self.scheduler.step()

    def on_epoch_begin(self, trainer):
        """Log at the beginning of train epoch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.

        """
        if not self.each_step:
            self.scheduler.step()

    def on_validation_end(self, trainer, val_loss):
        if self.best_current > val_loss:
            self.best_current = val_loss

        if self.scheduler.last_epoch >= self.Tmax:
            self.Tmax *= self.Tmult
            self.scheduler.last_epoch = -1
            self.scheduler.T_max = self.Tmax
            self.scheduler.base_lrs = [
                base_lr * self.lr_factor for base_lr in self.scheduler.base_lrs
            ]
            trainer.optimizer.load_state_dict(self.init_opt_state)

            if self.best_current >= self.best_previous:
                self.waiting += 1
            else:
                self.waiting = 0
                self.best_previous = self.best_current

            if self.waiting > self.patience:
                trainer._stop = True


class MaxEpochHook(Hook):
    """Hook to stop training when a maximum number of epochs is reached.

    Args:
       max_epochs (int): maximal number of epochs.

    """

    def __init__(self, max_epochs):
        self.max_epochs = max_epochs

    def on_epoch_begin(self, trainer):
        """Log at the beginning of train epoch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.

        """
        # stop training if max_epochs is reached
        if trainer.epoch > self.max_epochs:
            trainer._stop = True


class MaxStepHook(Hook):
    """Hook to stop training when a maximum number of steps is reached.

    Args:
        max_steps (int): maximum number of steps.

    """

    def __init__(self, max_steps):
        self.max_steps = max_steps

    def on_batch_begin(self, trainer, train_batch):
        """Log at the beginning of train batch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.
            train_batch (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        """
        # stop training if max_steps is reached
        if trainer.step > self.max_steps:
            trainer._stop = True


class LRScheduleHook(Hook):
    """Base class for learning rate scheduling hooks.

    This class provides a thin wrapper around torch.optim.lr_schedule._LRScheduler.

    Args:
        scheduler (torch.optim.lr_schedule._LRScheduler): scheduler.
        each_step (bool, optional): if set to True scheduler.step() is called every
            step, otherwise every epoch.

    """

    def __init__(self, scheduler, each_step=False):
        self.scheduler = scheduler
        self.each_step = each_step

    @property
    def state_dict(self):
        return {"scheduler": self.scheduler.state_dict()}

    @state_dict.setter
    def state_dict(self, state_dict):
        self.scheduler.load_state_dict(state_dict["scheduler"])

    def on_train_begin(self, trainer):
        self.scheduler.last_epoch = trainer.epoch - 1

    def on_batch_begin(self, trainer, train_batch):
        """Log at the beginning of train batch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.
            train_batch (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        """
        if self.each_step:
            self.scheduler.step()

    def on_epoch_begin(self, trainer):
        """Log at the beginning of train epoch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.

        """
        if not self.each_step:
            self.scheduler.step()


class ReduceLROnPlateauHook(Hook):
    r"""Hook for reduce plateau learning rate scheduling.

    This class provides a thin wrapper around
    torch.optim.lr_schedule.ReduceLROnPlateau. It takes the parameters
    of ReduceLROnPlateau as arguments and creates a scheduler from it whose
    step() function will be called every epoch.

    Args:
        patience (int, optional): number of epochs with no improvement after which
            learning rate will be reduced. For example, if `patience = 2`, then we
            will ignore the first 2 epochs with no improvement, and will only
            decrease the LR after the 3rd epoch if the loss still hasn't improved then.
        factor (float, optional): factor by which the learning rate will be reduced.
            new_lr = lr * factor.
        min_lr (float or list, optional): scalar or list of scalars. A lower bound on
            the learning rate of all param groups or each group respectively.
        window_length (int, optional): window over which the accumulated loss will
            be averaged.
        stop_after_min (bool, optional): if enabled stops after minimal learning rate
            is reached.

    """

    def __init__(
        self,
        optimizer,
        patience=25,
        factor=0.5,
        min_lr=1e-6,
        window_length=1,
        stop_after_min=False,
    ):
        self.patience = patience
        self.factor = factor
        self.min_lr = min_lr
        self.scheduler = ReduceLROnPlateau(
            optimizer, patience=self.patience, factor=self.factor, min_lr=self.min_lr
        )
        self.window_length = window_length
        self.stop_after_min = stop_after_min
        self.window = []

    @property
    def state_dict(self):
        return {"scheduler": self.scheduler.state_dict()}

    @state_dict.setter
    def state_dict(self, state_dict):
        self.scheduler.load_state_dict(state_dict["scheduler"])

    def on_validation_end(self, trainer, val_loss):
        self.window.append(val_loss)
        if len(self.window) > self.window_length:
            self.window.pop(0)
        accum_loss = np.mean(self.window)

        self.scheduler.step(accum_loss)

        if self.stop_after_min:
            for i, param_group in enumerate(self.scheduler.optimizer.param_groups):
                old_lr = float(param_group["lr"])
                if old_lr <= self.scheduler.min_lrs[i]:
                    trainer._stop = True


class ExponentialDecayHook(Hook):
    """Hook for reduce plateau learning rate scheduling.

    This class provides a thin wrapper around torch.optim.lr_schedule.StepLR.
    It takes the parameters of StepLR as arguments and creates a scheduler
    from it whose step() function will be called every
    step.

    Args:
        gamma (float): Factor by which the learning rate will be
            reduced. new_lr = lr * gamma
        step_size (int): Period of learning rate decay.

    """

    def __init__(self, optimizer, gamma=0.96, step_size=100000):
        self.scheduler = StepLR(optimizer, step_size, gamma)

    def on_batch_end(self, trainer, train_batch, result, loss):
        self.scheduler.step()


class UpdatePrioritiesHook(Hook):
    r"""Hook for updating the priority sampler"""

    def __init__(self, prioritized_sampler, priority_fn):
        self.prioritized_sampler = prioritized_sampler
        self.update_fn = priority_fn

    def on_batch_end(self, trainer, train_batch, result, loss):
        idx = train_batch["_idx"]
        self.prioritized_sampler.update_weights(
            idx.data.cpu().squeeze(),
            self.update_fn(train_batch, result).data.cpu().squeeze(),
        )



import os
import csv


__all__ = ["evaluate", "evaluate_dataset"]


def evaluate(
    args,
    model,
    train_loader,
    val_loader,
    test_loader,
    device,
    metrics,
    custom_header=None,
):

    header = []
    results = []

    loaders = dict(train=train_loader, validation=val_loader, test=test_loader)
    for datasplit in args.split:
        header += ["{} MAE".format(datasplit), "{} RMSE".format(datasplit)]
        derivative = model.output_modules[0].derivative
        if derivative is not None:
            header += [
                "{} MAE ({})".format(datasplit, derivative),
                "{} RMSE ({})".format(datasplit, derivative),
            ]
        results += evaluate_dataset(metrics, model, loaders[datasplit], device)

    if custom_header:
        header = custom_header

    eval_file = os.path.join(args.modelpath, "evaluation.txt")
    with open(eval_file, "w") as file:
        wr = csv.writer(file)
        wr.writerow(header)
        wr.writerow(results)


def evaluate_dataset(metrics, model, loader, device):
    model.eval()

    for metric in metrics:
        metric.reset()

    for batch in loader:
        batch = {k: v.to(device) for k, v in batch.items()}
        result = model(batch)

        for metric in metrics:
            metric.add_batch(batch, result)

    results = [metric.aggregate() for metric in metrics]
    return results



import json
import logging
from argparse import Namespace

import numpy as np
import torch
from ase.db import connect
from base64 import b64decode
from tqdm import tqdm

__all__ = [
    "set_random_seed",
    "count_params",
    "to_json",
    "read_from_json",
    "DeprecationHelper",
    "load_model",
    "read_deprecated_database",
    "activate_stress_computation",
]


def set_random_seed(seed):
    """
    This function sets the random seed (if given) or creates one for torch and numpy random state initialization

    Args:
        seed (int, optional): if seed not present, it is generated based on time
    """
    import time
    import numpy as np

    # 1) if seed not present, generate based on time
    if seed is None:
        seed = int(time.time() * 1000.0)
        # Reshuffle current time to get more different seeds within shorter time intervals
        # Taken from https://stackoverflow.com/questions/27276135/python-random-system-time-seed
        # & Gets overlapping bits, << and >> are binary right and left shifts
        seed = (
            ((seed & 0xFF000000) >> 24)
            + ((seed & 0x00FF0000) >> 8)
            + ((seed & 0x0000FF00) << 8)
            + ((seed & 0x000000FF) << 24)
        )
    # 2) Set seed for numpy (e.g. splitting)
    np.random.seed(seed)
    # 3) Set seed for torch (manual_seed now seeds all CUDA devices automatically)
    torch.manual_seed(seed)
    logging.info("Random state initialized with seed {:<10d}".format(seed))


def count_params(model):
    """
    This function takes a model as an input and returns the number of
    trainable parameters.

    Args:collect
        model (AtomisticModel): model for which you want to count
                                the trainable parameters

    Returns:
        params (int): number of trainable parameters for the model
    """
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    return params


def to_json(jsonpath, argparse_dict):
    """
    This function creates a .json file as a copy of argparse_dict

    Args:
        jsonpath (str): path to the .json file
        argparse_dict (dict): dictionary containing arguments from argument parser
    """
    with open(jsonpath, "w") as fp:
        json.dump(argparse_dict, fp, sort_keys=True, indent=4)


def read_from_json(jsonpath):
    """
    This function reads args from a .json file and returns the content as a namespace dict

    Args:
        jsonpath (str): path to the .json file

    Returns:
        namespace_dict (Namespace): namespace object build from the dict stored into the given .json file.
    """
    with open(jsonpath) as handle:
        dict = json.loads(handle.read())
        namespace_dict = Namespace(**dict)
    return namespace_dict


class DeprecationHelper(object):
    def __init__(self, new_target, old_name):
        self.new_target = new_target
        self.old_name = old_name

    def _warn(self):
        from warnings import warn

        warn(
            self.old_name
            + "is deprecated, use "
            + self.new_target.__class__.__name__
            + " instead",
            DeprecationWarning,
        )

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self.new_target, attr)


def load_model(model_path, map_location=None):
    """
    Wrapper function for `for safely loading models where certain new attributes of the model class are not present.
    E.g. "requires_stress" for computing the stress tensor.

    Args:
        model_path (str): Path to the model file.
        map_location (torch.device): Device where the model should be loaded to.

    Returns:
        :class:`schnetpack.atomistic.AtomisticModel`: Loaded SchNetPack model.

    """
    model = torch.load(model_path, map_location=map_location)

    # Check for data parallel models
    if hasattr(model, "module"):
        model_module = model.module
        output_modules = model.module.output_modules
    else:
        model_module = model
        output_modules = model.output_modules

    # Set stress tensor attribute if not present
    if not hasattr(model_module, "requires_stress"):
        model_module.requires_stress = False
        for module in output_modules:
            module.stress = None

    return model


def read_deprecated_database(db_path):
    """
    Read all atoms and properties from deprecated ase databases.

    Args:
        db_path (str): Path to deprecated database

    Returns:
        atoms (list): All atoms objects of the database.
        properties (list): All property dictionaries of the database.

    """
    with connect(db_path) as conn:
        db_size = conn.count()
    atoms = []
    properties = []
    key_value_pairs = []

    for idx in tqdm(range(1, db_size + 1), "Reading deprecated database"):
        with connect(db_path) as conn:
            row = conn.get(idx)

        at = row.toatoms()
        pnames = [pname for pname in row.data.keys() if not pname.startswith("_")]
        props = {}
        for pname in pnames:
            try:
                shape = row.data["_shape_" + pname]
                dtype = row.data["_dtype_" + pname]
                prop = np.frombuffer(b64decode(row.data[pname]), dtype=dtype)
                prop = prop.reshape(shape)
            except:
                # fallback for properties stored directly
                # in the row
                if pname in row:
                    prop = row[pname]
                else:
                    prop = row.data[pname]

                try:
                    prop.shape
                except AttributeError as e:
                    prop = np.array([prop], dtype=np.float32)
            props[pname] = prop

        atoms.append(at)
        properties.append(props)
        key_value_pairs.append(row.key_value_pairs)

    return atoms, properties, key_value_pairs


def activate_stress_computation(model, stress="stress"):
    """
    Utility function to activate the computation of the stress tensor for a model not trained explicitly on
    this property. It is recommended to at least have used forces during training when switching on the stress.
    Moreover, now proper crystal cell (volume > 0) needs to be specified for the molecules.

    Args:
        model (schnetpack.atomistic.AtomisticModel): SchNetPack model for which computation of the stress tensor
                                                    should be activated.
        stress (str): Designated name of the stress tensor property used in the model output.
    """
    # Check for data parallel models
    if hasattr(model, "module"):
        model_module = model.module
        output_modules = model.module.output_modules
    else:
        model_module = model
        output_modules = model.output_modules

    # Set stress tensor attribute if not present
    if hasattr(model_module, "requires_stress"):
        model_module.requires_stress = True
        for module in output_modules:
            if hasattr(module, "stress"):
                module.stress = stress



import os

import numpy as np
from ase import Atoms
from ase.db import connect
from ase.units import eV

import schnetpack as spk
from schnetpack.data import AtomsDataError, AtomsDataSubset
from schnetpack.datasets import DownloadableAtomsData

__all__ = ["MaterialsProject"]


class MaterialsProject(DownloadableAtomsData):
    """Materials Project (MP) database of bulk crystals.

    This class adds convenient functions to download Materials Project data into
    pytorch.

    Args:
        dbpath (str): path to directory containing database.
        cutoff (float): cutoff for bulk interactions.
        apikey (str, optional): materials project key needed to download the data.
        download (bool, optional): enable downloading if database does not exists.
        subset (list, optional): Deprecated! Do not use! Subsets are created with
            AtomsDataSubset class.
        load_only (list, optional): reduced set of properties to be loaded
        collect_triples (bool, optional): Set to True if angular features are needed.
        environment_provider (spk.environment.BaseEnvironmentProvider): define how
            neighborhood is calculated
            (default=spk.environment.SimpleEnvironmentProvider).

    """

    # properties
    EformationPerAtom = "formation_energy_per_atom"
    EPerAtom = "energy_per_atom"
    BandGap = "band_gap"
    TotalMagnetization = "total_magnetization"

    def __init__(
        self,
        dbpath,
        apikey=None,
        download=True,
        subset=None,
        load_only=None,
        collect_triples=False,
        environment_provider=spk.environment.SimpleEnvironmentProvider(),
    ):

        available_properties = [
            MaterialsProject.EformationPerAtom,
            MaterialsProject.EPerAtom,
            MaterialsProject.BandGap,
            MaterialsProject.TotalMagnetization,
        ]

        units = [eV, eV, eV, 1.0]

        self.apikey = apikey

        super(MaterialsProject, self).__init__(
            dbpath=dbpath,
            subset=subset,
            load_only=load_only,
            environment_provider=environment_provider,
            collect_triples=collect_triples,
            available_properties=available_properties,
            units=units,
            download=download,
        )

    def create_subset(self, idx):
        idx = np.array(idx)
        subidx = idx if self.subset is None else np.array(self.subset)[idx]

        return MaterialsProject(
            dbpath=self.dbpath,
            download=False,
            subset=subidx,
            load_only=self.load_only,
            collect_triples=self.collect_triples,
            environment_provider=self.environment_provider,
        )

    def at_timestamp(self, timestamp):
        """
        Returns a new dataset that only consists of items created before
        the given timestamp.

        Args:
            timestamp (str): timestamp

        Returns:
            schnetpack.datasets.matproj.MaterialsProject: dataset with subset of
                original data
        """
        with connect(self.dbpath) as conn:
            rows = conn.select(columns=["id", "key_value_pairs"])
            idxs = []
            timestamps = []
            for row in rows:
                idxs.append(row.id - 1)
                timestamps.append(row.key_value_pairs["created_at"])
        idxs = np.array(idxs)
        timestamps = np.array(timestamps)
        return AtomsDataSubset(self, idxs[timestamps <= timestamp])

    def _download(self):
        """
        Downloads dataset provided it does not exist in self.path

        Returns:
            works (bool): true if download succeeded or file already exists
        """
        if self.apikey is None:
            raise AtomsDataError(
                "Provide a valid API key in order to download the "
                "Materials Project data!"
            )
        try:
            from pymatgen.ext.matproj import MPRester
            from pymatgen.core import Structure
            import pymatgen as pmg
        except:
            raise ImportError(
                "In order to download Materials Project data, you have to install "
                "pymatgen"
            )

        # collect data
        atms_list = []
        properties_list = []
        key_value_pairs_list = []
        with MPRester(self.apikey) as m:
            for N in range(1, 9):
                for nsites in range(0, 300, 30):
                    ns = {"$lt": nsites + 31, "$gt": nsites}
                    query = m.query(
                        criteria={
                            "nelements": N,
                            "is_compatible": True,
                            "nsites": ns,
                        },
                        properties=[
                            "structure",
                            "energy_per_atom",
                            "formation_energy_per_atom",
                            "total_magnetization",
                            "band_gap",
                            "material_id",
                            "warnings",
                            "created_at",
                        ],
                    )
                    for k, q in enumerate(query):
                        s = q["structure"]
                        if type(s) is Structure:
                            atms_list.append(
                                Atoms(
                                    numbers=s.atomic_numbers,
                                    positions=s.cart_coords,
                                    cell=s.lattice.matrix,
                                    pbc=True,
                                )
                            )
                            properties_list.append(
                                {
                                    MaterialsProject.EPerAtom: q["energy_per_atom"],
                                    MaterialsProject.EformationPerAtom: q[
                                        "formation_energy_per_atom"
                                    ],
                                    MaterialsProject.TotalMagnetization: q[
                                        "total_magnetization"
                                    ],
                                    MaterialsProject.BandGap: q["band_gap"],
                                }
                            )
                            key_value_pairs_list.append(
                                {
                                    "material_id": q["material_id"],
                                    "created_at": q["created_at"],
                                }
                            )

        # write systems to database
        self.add_systems(
            atms_list,
            property_list=properties_list,
            key_value_pairs_list=key_value_pairs_list,
        )

        self.set_metadata({})



"""
This module contains various thermostats for regulating the temperature of the system during
molecular dynamics simulations. Apart from standard thermostats for convetional simulations,
a series of special thermostats developed for ring polymer molecular dynamics is also provided.
"""
import torch
import numpy as np
import scipy.linalg as linalg
import logging

from schnetpack.md.utils import (
    MDUnits,
    load_gle_matrices,
    NormalModeTransformer,
    YSWeights,
)
from schnetpack.md.integrators import RingPolymer
from schnetpack.md.simulation_hooks.basic_hooks import SimulationHook

__all__ = [
    "ThermostatHook",
    "BerendsenThermostat",
    "GLEThermostat",
    "PIGLETThermostat",
    "LangevinThermostat",
    "PILELocalThermostat",
    "PILEGlobalThermostat",
    "NHCThermostat",
    "NHCRingPolymerThermostat",
    "TRPMDThermostat",
]


class ThermostatError(Exception):
    """
    Exception for thermostat class.
    """

    pass


class ThermostatHook(SimulationHook):
    """
    Basic thermostat hook for simulator class. This class is initialized based on the simulator and system
    specifications during the first MD step. Thermostats are applied before and after each MD step.

    Args:
        temperature_bath (float): Temperature of the heat bath in Kelvin.
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.
        detach (bool): Whether the computational graph should be detached after each simulation step. Default is true,
                       should be changed if differentiable MD is desired.
    """

    # TODO: Could be made a torch nn.Module

    def __init__(self, temperature_bath, nm_transformation=None, detach=True):
        self.temperature_bath = temperature_bath
        self.initialized = False
        self.device = None
        self.n_replicas = None
        self.nm_transformation = nm_transformation
        self.detach = detach

    def on_simulation_start(self, simulator):
        """
        Routine to initialize the thermostat based on the current state of the simulator. Reads the device to be uses,
        as well as the number of molecular replicas present in simulator.system. Furthermore, the normal mode
        transformer is initialized during ring polymer simulations. In addition, a flag is set so that the thermostat
        is not reinitialized upon continuation of the MD.

        Main function is the _init_thermostat routine, which takes the simulator as input and must be provided for every
        new thermostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        self.device = simulator.system.device
        self.n_replicas = simulator.system.n_replicas

        # Check if using normal modes is feasible and initialize
        if self.nm_transformation is not None:
            if not isinstance(simulator.integrator, RingPolymer):
                raise ThermostatError(
                    "Normal mode transformation should only"
                    "be used with ring polymer dynamics."
                )
            else:
                # If simulation is not restarted from a previous point, initialize.
                if simulator.effective_steps == 0:
                    self.nm_transformation = self.nm_transformation(
                        self.n_replicas, device=self.device
                    )

        if not self.initialized:
            self._init_thermostat(simulator)
            self.initialized = True

    def on_step_begin(self, simulator):
        """
        First application of the thermostat befor the first half step of the dynamics. Regulates temperature and applies
        a mask to the system momenta in order to avoid problems of e.g. thermal noise added to the zero padded tensors.
        The detach is carried out here.

        Main function is the _apply_thermostat routine, which takes the simulator as input and must be provided for
        every new thermostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Apply thermostat
        self._apply_thermostat(simulator)

        # Re-apply atom masks for differently sized molecules, as some
        # thermostats add random noise
        simulator.system.momenta = (
            simulator.system.momenta * simulator.system.atom_masks
        )

        # Detach if requested
        if self.detach:
            simulator.system.momenta = simulator.system.momenta.detach()

    def on_step_end(self, simulator):
        """
        First application of the thermostat befor the first half step of the dynamics. Regulates temperature and applies
        a mask to the system momenta in order to avoid problems of e.g. thermal noise added to the zero padded tensors.
        The detach is carried out here.

        Main function is the _apply_thermostat routine, which takes the simulator as input and must be provided for
        every new thermostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Apply thermostat
        self._apply_thermostat(simulator)

        # Re-apply atom masks for differently sized molecules, as some
        # thermostats add random noise
        simulator.system.momenta = (
            simulator.system.momenta * simulator.system.atom_masks
        )

        # Detach if requested
        if self.detach:
            simulator.system.momenta = simulator.system.momenta.detach()

    def _init_thermostat(self, simulator):
        """
        Dummy routine for initializing a thermostat based on the current simulator. Should be implemented for every
        new thermostat. Has access to the information contained in the simulator class, e.g. number of replicas, time
        step, masses of the atoms, etc.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        pass

    def _apply_thermostat(self, simulator):
        """
        Dummy routine for applying the thermostat to the system. Should use the implemented thermostat to update the
        momenta of the system contained in simulator.system.momenta. Is called twice each simulation time step.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        raise NotImplementedError


class BerendsenThermostat(ThermostatHook):
    """
    Berendsen velocity rescaling thermostat, as described in [#berendsen1]_. Simple thermostat for e.g. equilibrating
    the system, does not sample the canonical ensemble.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs

    References
    ----------
    .. [#berendsen1] Berendsen, Postma, van Gunsteren, DiNola, Haak:
       Molecular dynamics with coupling to an external bath.
       The Journal of Chemical Physics, 81 (8), 3684-3690. 1984.
    """

    def __init__(self, temperature_bath, time_constant):
        super(BerendsenThermostat, self).__init__(temperature_bath)

        self.time_constant = time_constant * MDUnits.fs2internal

    def _apply_thermostat(self, simulator):
        """
        Apply the Berendsen thermostat via rescaling the systems momenta based on the current instantaneous temperature
        and the bath temperature.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        scaling = 1.0 + simulator.integrator.time_step / self.time_constant * (
            self.temperature_bath / simulator.system.temperature - 1
        )
        simulator.system.momenta = (
            torch.sqrt(scaling[:, :, None, None]) * simulator.system.momenta
        )


class GLEThermostat(ThermostatHook):
    """
    Stochastic generalized Langevin colored noise thermostat by Ceriotti et. al. as described in [#gle_thermostat1]_.
    This thermostat requires specially parametrized matrices, which can be obtained online from:
    http://gle4md.org/index.html?page=matrix

    The additional degrees of freedom added to the system are defined via the matrix dimensions. This could in principle
    be used for ring polymer dynamics by providing a normal mode transformation.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        gle_file (str): File containing the GLE matrices
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.

    References
    ----------
    .. [#gle_thermostat1] Ceriotti, Bussi, Parrinello:
       Colored-noise thermostats à la carte.
       Journal of Chemical Theory and Computation 6 (4), 1170-1180. 2010.
    """

    def __init__(self, temperature_bath, gle_file, nm_transformation=None):
        super(GLEThermostat, self).__init__(
            temperature_bath, nm_transformation=nm_transformation
        )

        self.gle_file = gle_file

        # To be initialized on beginning of the simulation, once system and
        # integrator are known
        self.c1 = None
        self.c2 = None
        self.thermostat_momenta = None
        self.thermostat_factor = None

    def _init_thermostat(self, simulator):
        """
        Initialize the GLE thermostat by reading in the the required matrices and setting up the initial random
        thermostat momenta and the mass factor.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Generate main matrices
        self.c1, self.c2 = self._init_gle_matrices(simulator)

        # Get particle masses
        self.thermostat_factor = torch.sqrt(simulator.system.masses)[..., None]

        # Get initial thermostat momenta
        self.thermostat_momenta = self._init_thermostat_momenta(simulator)

    def _init_gle_matrices(self, simulator):
        """
        Read all GLE matrices from a file and check, whether they have the right dimensions.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        a_matrix, c_matrix = load_gle_matrices(self.gle_file)

        if a_matrix is None:
            raise ThermostatError(
                "Error reading GLE matrices" " from {:s}".format(self.gle_file)
            )
        elif a_matrix.shape[0] > 1:
            raise ThermostatError(
                "More than one A matrix found. Could be " "PIGLET input."
            )
        else:
            # Remove leading dimension (for normal modes)
            a_matrix = a_matrix.squeeze()

        c1, c2 = self._init_single_gle_matrix(a_matrix, c_matrix, simulator)
        return c1, c2

    def _init_single_gle_matrix(self, a_matrix, c_matrix, simulator):
        """
        Based on the matrices found in the GLE file, initialize the GLE matrices required for a simulation with the
        thermostat. See [#stochastic_thermostats1]_ for more detail. The dimensions of all matrices are:
        degrees_of_freedom x degrees_of_freedom,
        where degrees_of_freedom are the degrees of freedom of the extended system.

        Args:
            a_matrix (np.array): Raw matrices containing friction friction acting on system (drift matrix).
            c_matrix (np.array): Raw matrices modulating the intensity of the random force (diffusion matrix).
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.

        Returns:
            torch.Tensor: Drift matrix for simulation.
            torch.Tensor: Diffusion matrix initialized for simulation.

        References
        ----------
        .. [#stochastic_thermostats1]_Ceriotti, Parrinello, Markland, Manolopoulos:
           Efficient stochastic thermostatting of path integral molecular dynamics.
           The Journal of Chemical Physics, 133 (12), 124104. 2010.
        """
        if c_matrix is None:
            c_matrix = np.eye(a_matrix.shape[-1]) * self.temperature_bath * MDUnits.kB
            # Check if normal GLE or GLE for ring polymers is needed:
            if isinstance(simulator.integrator, RingPolymer):
                logging.info(
                    "RingPolymer integrator detected, initializing " "C accordingly."
                )
                c_matrix *= simulator.system.n_replicas
        else:
            c_matrix = c_matrix.squeeze()
            logging.info(
                "C matrix for GLE loaded, provided temperature will " "be ignored."
            )

        # A does not need to be transposed, else c2 is imaginary
        c1 = linalg.expm(-0.5 * simulator.integrator.time_step * a_matrix)

        # c2 is symmetric
        c2 = linalg.sqrtm(c_matrix - np.dot(c1, np.dot(c_matrix, c1.T)))

        # To myself: original expression is c1 = exp(-dt/2 * a.T)
        # the C1 here is c1.T, since exp(-dt/2*a.T).T = exp(-dt/2*a)
        # The formula for c2 is c2 = sqrtm(1-c1.T*c1)
        # In our case, this becomes sqrtm(1-C1*C1.T)
        # For the propagation we have the original expression c1*p, where
        # p is a column vector (ndegrees x something)
        # In our case P is (something x ndegrees), hence p.T
        # The propagation then becomes P*C1 = p.T*c1.T = (c1*p).T
        # c2 is symmetric by construction, hence C2=c2
        c1 = torch.from_numpy(c1).to(self.device).float()
        c2 = torch.from_numpy(c2).to(self.device).float()
        return c1, c2

    def _init_thermostat_momenta(self, simulator, free_particle_limit=True):
        """
        Initialize the thermostat momenta tensor based on the system specifications. This tensor is then updated
        during the GLE dynamics.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
            free_particle_limit (bool): Initialize momenta according to free particle limit instead of a zero matrix
                                        (default=True).

        Returns:
            torch.Tensor: Initialized random momenta of the extended system with the dimension:
                          n_replicas x n_molecules x n_atoms x 3 x degrees_of_freedom
        """
        degrees_of_freedom = self.c1.shape[-1]
        if not free_particle_limit:
            initial_momenta = torch.zeros(
                *simulator.system.momenta.shape, degrees_of_freedom, device=self.device
            )
        else:
            initial_momenta = torch.randn(
                *simulator.system.momenta.shape, degrees_of_freedom, device=self.device
            )
            initial_momenta = torch.matmul(initial_momenta, self.c2)
        return initial_momenta

    def _apply_thermostat(self, simulator):
        """
        Perform the update of the system momenta according to the GLE thermostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Generate random noise
        thermostat_noise = torch.randn(
            self.thermostat_momenta.shape, device=self.device
        )

        # Get current momenta
        momenta = simulator.system.momenta

        # Apply transformation if requested
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.beads2normal(momenta)

        # Set current momenta
        self.thermostat_momenta[:, :, :, :, 0] = momenta

        # Apply thermostat
        self.thermostat_momenta = (
            torch.matmul(self.thermostat_momenta, self.c1)
            + torch.matmul(thermostat_noise, self.c2) * self.thermostat_factor
        )

        # Extract momenta
        momenta = self.thermostat_momenta[:, :, :, :, 0]

        # Apply transformation if requested
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.normal2beads(momenta)

        simulator.system.momenta = momenta

    @property
    def state_dict(self):
        state_dict = {
            "c1": self.c1,
            "c2": self.c2,
            "thermostat_factor": self.thermostat_factor,
            "thermostat_momenta": self.thermostat_momenta,
            "temperature_bath": self.temperature_bath,
            "n_replicas": self.n_replicas,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.c1 = state_dict["c1"]
        self.c2 = state_dict["c2"]
        self.thermostat_factor = state_dict["thermostat_factor"]
        self.thermostat_momenta = state_dict["thermostat_momenta"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.n_replicas = state_dict["n_replicas"]

        # Set initialized flag
        self.initialized = True


class PIGLETThermostat(GLEThermostat):
    """
    Efficient generalized Langevin equation stochastic thermostat for ring polymer dynamics simulations, see
    [#piglet_thermostat1]_ for a detailed description. In contrast to the standard GLE thermostat, every normal mode
    of the ring polymer is
    thermostated seperately.


    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        gle_file (str): File containing the GLE matrices
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.

    References
    ----------
    .. [#piglet_thermostat1] Uhl, Marx, Ceriotti:
       Accelerated path integral methods for atomistic simulations at ultra-low temperatures.
       The Journal of chemical physics, 145(5), 054101. 2016.
    """

    def __init__(
        self, temperature_bath, gle_file, nm_transformation=NormalModeTransformer
    ):

        logging.info("Using PIGLET thermostat")
        super(PIGLETThermostat, self).__init__(
            temperature_bath, gle_file, nm_transformation=nm_transformation
        )

    def _init_gle_matrices(self, simulator):
        """
        Initialize the matrices necessary for the PIGLET thermostat. In contrast to the basic GLE thermostat, these
        have the dimension:
        n_replicas x degrees_of_freedom x degrees_of_freedom,
        where n_replicas is the number of beads in the ring polymer and degrees_of_freedom is the number of degrees of
        freedom introduced by GLE.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.

        Returns:
            torch.Tensor: Drift matrices for the PIGLET thermostat.
            torch.Tensor: Diffusion matrices.
        """
        a_matrix, c_matrix = load_gle_matrices(self.gle_file)

        if a_matrix is None:
            raise ThermostatError(
                "Error reading GLE matrices " "from {:s}".format(self.gle_file)
            )
        if a_matrix.shape[0] != self.n_replicas:
            raise ThermostatError(
                "Expected {:d} beads but "
                "found {:d}.".format(self.n_replicas, a_matrix.shape[0])
            )

        if not isinstance(simulator.integrator, RingPolymer):
            raise ThermostatError("PIGLET thermostat should only be used with " "RPMD.")

        all_c1 = []
        all_c2 = []

        # Generate main matrices
        for b in range(self.n_replicas):
            c1, c2 = self._init_single_gle_matrix(
                a_matrix[b], (c_matrix[b], None)[c_matrix is None], simulator
            )
            # Add extra dimension for use with torch.cat, correspond to normal
            # modes of ring polymer
            all_c1.append(c1[None, ...])
            all_c2.append(c2[None, ...])

        # Bring to correct shape for later matmul broadcasting
        c1 = torch.cat(all_c1)[:, None, None, :, :]
        c2 = torch.cat(all_c2)[:, None, None, :, :]
        return c1, c2


class LangevinThermostat(ThermostatHook):
    """
    Basic stochastic Langevin thermostat, see e.g. [#langevin_thermostat1]_ for more details.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.

    References
    ----------
    .. [#langevin_thermostat1] Bussi, Parrinello:
       Accurate sampling using Langevin dynamics.
       Physical Review E, 75(5), 056707. 2007.
    """

    def __init__(self, temperature_bath, time_constant, nm_transformation=None):

        logging.info("Using Langevin thermostat")
        super(LangevinThermostat, self).__init__(
            temperature_bath, nm_transformation=nm_transformation
        )

        self.time_constant = time_constant * MDUnits.fs2internal

        self.thermostat_factor = None
        self.c1 = None
        self.c2 = None

    def _init_thermostat(self, simulator):
        """
        Initialize the Langevin coefficient matrices based on the system and simulator properties.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Initialize friction coefficients
        gamma = torch.ones(1, device=self.device) / self.time_constant

        # Initialize coefficient matrices
        c1 = torch.exp(-0.5 * simulator.integrator.time_step * gamma)
        c2 = torch.sqrt(1 - c1 ** 2)

        self.c1 = c1.to(self.device)[:, None, None, None]
        self.c2 = c2.to(self.device)[:, None, None, None]

        # Get mass and temperature factors
        self.thermostat_factor = torch.sqrt(
            simulator.system.masses * MDUnits.kB * self.temperature_bath
        )

    def _apply_thermostat(self, simulator):
        """
        Apply the stochastic Langevin thermostat to the systems momenta.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Get current momenta
        momenta = simulator.system.momenta

        # Apply transformation
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.beads2normal(momenta)

        # Generate random noise
        thermostat_noise = torch.randn(momenta.shape, device=self.device)

        # Apply thermostat
        momenta = (
            self.c1 * momenta + self.thermostat_factor * self.c2 * thermostat_noise
        )

        # Apply transformation if requested
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.normal2beads(momenta)

        simulator.system.momenta = momenta

    @property
    def state_dict(self):
        state_dict = {
            "c1": self.c1,
            "c2": self.c2,
            "thermostat_factor": self.thermostat_factor,
            "temperature_bath": self.temperature_bath,
            "n_replicas": self.n_replicas,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.c1 = state_dict["c1"]
        self.c2 = state_dict["c2"]
        self.thermostat_factor = state_dict["thermostat_factor"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.n_replicas = state_dict["n_replicas"]

        # Set initialized flag
        self.initialized = True


class PILELocalThermostat(LangevinThermostat):
    """
    Langevin thermostat for ring polymer molecular dynamics as introduced in [#stochastic_thermostats2]_.
    Applies specially initialized Langevin thermostats to the beads of the ring polymer in normal mode representation.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.
        thermostat_centroid (bool): Whether a thermostat should be applied to the centroid of the ring polymer in
                                    normal mode representation (relevant e.g. for TRPMD, default is True)
        damping (float): If specified, damping factor is applied to the current momenta of the system (used in TRPMD,
                         default is no damping).

    References
    ----------
    .. [#stochastic_thermostats2] Ceriotti, Parrinello, Markland, Manolopoulos:
       Efficient stochastic thermostatting of path integral molecular dynamics.
       The Journal of Chemical Physics, 133 (12), 124104. 2010.
    """

    def __init__(
        self,
        temperature_bath,
        time_constant,
        nm_transformation=NormalModeTransformer,
        thermostat_centroid=True,
        damping=None,
    ):
        super(PILELocalThermostat, self).__init__(
            temperature_bath, time_constant, nm_transformation=nm_transformation
        )
        self.thermostat_centroid = thermostat_centroid
        self.damping = damping

    def _init_thermostat(self, simulator):
        """
        Initialize the Langevin matrices based on the normal mode frequencies of the ring polymer. If the centroid is to
        be thermostatted, the suggested value of 1/time_constant is used.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        if not isinstance(simulator.integrator, RingPolymer):
            raise ThermostatError("PILE thermostats can only be used in RPMD")

        # Initialize friction coefficients
        gamma_normal = 2 * simulator.integrator.omega_normal

        # Use seperate coefficient for centroid mode (default, unless using thermostatted RPMD)
        if self.thermostat_centroid:
            gamma_normal[0] = 1.0 / self.time_constant

        # Apply TRPMD damping factor if provided
        if self.damping is not None:
            gamma_normal *= self.damping

        if self.nm_transformation is None:
            raise ThermostatError(
                "Normal mode transformation required for " "PILE thermostat"
            )

        # Initialize coefficient matrices
        c1 = torch.exp(-0.5 * simulator.integrator.time_step * gamma_normal)
        c2 = torch.sqrt(1 - c1 ** 2)

        self.c1 = c1.to(self.device)[:, None, None, None]
        self.c2 = c2.to(self.device)[:, None, None, None]

        # Get mass and temperature factors
        self.thermostat_factor = torch.sqrt(
            simulator.system.masses
            * MDUnits.kB
            * self.n_replicas
            * self.temperature_bath
        )

    @property
    def state_dict(self):
        state_dict = {
            "c1": self.c1,
            "c2": self.c2,
            "thermostat_factor": self.thermostat_factor,
            "temperature_bath": self.temperature_bath,
            "n_replicas": self.n_replicas,
            "damping": self.damping,
            "thermostat_centroid": self.thermostat_centroid,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.c1 = state_dict["c1"]
        self.c2 = state_dict["c2"]
        self.thermostat_factor = state_dict["thermostat_factor"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.n_replicas = state_dict["n_replicas"]
        self.damping = state_dict["damping"]
        self.thermostat_centroid = state_dict["thermostat_centroid"]

        # Set initialized flag
        self.initialized = True


class PILEGlobalThermostat(PILELocalThermostat):
    """
    Global variant of the ring polymer Langevin thermostat as suggested in [#stochastic_thermostats3]_. This thermostat
    applies a stochastic velocity rescaling thermostat [#stochastic_velocity_rescaling1]_ to the ring polymer centroid
    in normal mode representation.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.

    References
    ----------
    .. [#stochastic_thermostats3] Ceriotti, Parrinello, Markland, Manolopoulos:
       Efficient stochastic thermostatting of path integral molecular dynamics.
       The Journal of Chemical Physics, 133 (12), 124104. 2010.
    .. [#stochastic_velocity_rescaling1] Bussi, Donadio, Parrinello:
       Canonical sampling through velocity rescaling.
       The Journal of chemical physics, 126(1), 014101. 2007.
    """

    def __init__(
        self, temperature_bath, time_constant, nm_transformation=NormalModeTransformer
    ):
        logging.info("Using global PILE thermostat")
        super(PILEGlobalThermostat, self).__init__(
            temperature_bath, time_constant, nm_transformation=nm_transformation
        )

    def _apply_thermostat(self, simulator):
        """
        Apply the global PILE thermostat to the system momenta. This is essentially the same as for the basic Langevin
        thermostat, with exception of replacing the equations for the centroid (index 0 in first dimension) with the
        stochastic velocity rescaling equations.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Get current momenta
        momenta = simulator.system.momenta

        # Apply transformation
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.beads2normal(momenta)

        # Generate random noise
        thermostat_noise = torch.randn(momenta.shape, device=self.device)

        # Apply thermostat to centroid mode
        c1_centroid = self.c1[0]

        momenta_centroid = momenta[0]
        thermostat_noise_centroid = thermostat_noise[0]

        # Compute kinetic energy of centroid
        kinetic_energy_factor = torch.sum(
            momenta_centroid ** 2 / simulator.system.masses[0]
        ) / (self.temperature_bath * MDUnits.kB * self.n_replicas)

        centroid_factor = (1 - c1_centroid) / kinetic_energy_factor

        alpha_sq = (
            c1_centroid
            + torch.sum(thermostat_noise_centroid ** 2) * centroid_factor
            + 2
            * thermostat_noise_centroid[0, 0, 0]
            * torch.sqrt(c1_centroid * centroid_factor)
        )

        alpha_sign = torch.sign(
            thermostat_noise_centroid[0, 0, 0]
            + torch.sqrt(c1_centroid / centroid_factor)
        )

        alpha = torch.sqrt(alpha_sq) * alpha_sign

        # Finally apply thermostat...
        momenta[0] = alpha * momenta[0]

        # Apply thermostat for remaining normal modes
        momenta[1:] = (
            self.c1[1:] * momenta[1:]
            + self.thermostat_factor * self.c2[1:] * thermostat_noise[1:]
        )

        # Apply transformation if requested
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.normal2beads(momenta)

        simulator.system.momenta = momenta


class TRPMDThermostat(PILELocalThermostat):
    """
    Thermostatted ring polymer molecular dynamics thermostat variant of the local PILE thermostat as introduced in
    [#trpmd_thermostat1]_. Here, no thermostat is applied to the centroid and the dynamics of the system are damped via
    a given damping factor.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        damping (float): Damping factor of the thermostat.
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module use dto transform between beads and normal
                                                                       model representation in ring polymer dynamics.

    References
    ----------
    .. [#trpmd_thermostat1] Rossi, Ceriotti, Manolopoulos:
       How to remove the spurious resonances from ring polymer molecular dynamics.
       The Journal of Chemical Physics, 140(23), 234116. 2014.
    """

    def __init__(
        self, temperature_bath, damping, nm_transformation=NormalModeTransformer
    ):
        super(TRPMDThermostat, self).__init__(
            temperature_bath,
            1.0,
            nm_transformation=nm_transformation,
            thermostat_centroid=False,
            damping=damping,
        )


class NHCThermostat(ThermostatHook):
    """
    Nose-Hover chain thermostat, which links the system to a chain of deterministic Nose-Hoover thermostats first
    introduced in [#nhc_thermostat1]_ and described in great detail in [#nhc_thermostat2]_. Advantage of the NHC
    thermostat is, that it does not apply random perturbations to the system and is hence fully deterministic. However,
    this comes at an increased numerical cost compared to e.g. the stochastic thermostats described above.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        chain_length (int): Number of Nose-Hoover thermostats applied in the chain.
        massive (bool): If set to true, an individual thermostat is applied to each degree of freedom in the system.
                        Can e.g. be used for thermostatting (default=False).
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module used to transform between beads and normal
                                                                   model representation in ring polymer dynamics.
        multi_step (int): Number of steps used for integrating the NH equations of motion (default=2)
        integration_order (int): Order of the Yoshida-Suzuki integrator used for propagating the thermostat (default=3).

    References
    ----------
    .. [#nhc_thermostat1] Tobias, Martyna, Klein:
       Molecular dynamics simulations of a protein in the canonical ensemble.
       The Journal of Physical Chemistry, 97(49), 12959-12966. 1993.
    .. [#nhc_thermostat2] Martyna, Tuckerman, Tobias, Klein:
       Explicit reversible integrators for extended systems dynamics.
       Molecular Physics, 87(5), 1117-1157. 1996.
    """

    def __init__(
        self,
        temperature_bath,
        time_constant,
        chain_length=3,
        massive=False,
        nm_transformation=None,
        multi_step=1,
        integration_order=3,
    ):
        super(NHCThermostat, self).__init__(
            temperature_bath, nm_transformation=nm_transformation
        )

        self.chain_length = chain_length
        self.massive = massive
        self.frequency = 1 / (time_constant * MDUnits.fs2internal)

        # Cpmpute kBT, since it will be used a lot
        self.kb_temperature = self.temperature_bath * MDUnits.kB

        # Propagation parameters
        self.multi_step = multi_step
        self.integration_order = integration_order
        self.time_step = None

        # Find out number of particles (depends on whether massive or not)
        self.degrees_of_freedom = None
        self.masses = None

        self.velocities = None
        self.positions = None
        self.forces = None

    def _init_thermostat(self, simulator):
        """
        Initialize the thermostat positions, forces, velocities and masses, as well as the number of degrees of freedom
        seen by each chain link.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Determine integration step via multi step and Yoshida Suzuki weights
        integration_weights = YSWeights(self.device).get_weights(self.integration_order)
        self.time_step = (
            simulator.integrator.time_step * integration_weights / self.multi_step
        )

        # Determine shape of tensors and internal degrees of freedom
        n_replicas, n_molecules, n_atoms, xyz = simulator.system.momenta.shape

        if self.massive:
            state_dimension = (n_replicas, n_molecules, n_atoms, xyz, self.chain_length)
            # Since momenta will be masked later, no need to set non-atoms to 0
            self.degrees_of_freedom = torch.ones(
                (n_replicas, n_molecules, n_atoms, xyz), device=self.device
            )
        else:
            state_dimension = (n_replicas, n_molecules, 1, 1, self.chain_length)
            self.degrees_of_freedom = (
                3 * simulator.system.n_atoms.float()[None, :, None, None]
            )

        # Set up masses
        self._init_masses(state_dimension, simulator)

        # Set up internal variables
        self.positions = torch.zeros(state_dimension, device=self.device)
        self.forces = torch.zeros(state_dimension, device=self.device)
        self.velocities = torch.zeros(state_dimension, device=self.device)

    def _init_masses(self, state_dimension, simulator):
        """
        Auxiliary routine for initializing the thermostat masses.

        Args:
            state_dimension (tuple): Size of the thermostat states. This is used to differentiate between the massive
                                     and the standard algorithm
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        self.masses = torch.ones(state_dimension, device=self.device)
        # Get masses of innermost thermostat
        self.masses[..., 0] = (
            self.degrees_of_freedom * self.kb_temperature / self.frequency ** 2
        )
        # Set masses of remaining thermostats
        self.masses[..., 1:] = self.kb_temperature / self.frequency ** 2

    def _propagate_thermostat(self, kinetic_energy):
        """
        Propagation step of the NHC thermostat. Please refer to [#nhc_thermostat2]_ for more detail on the algorithm.

        Args:
            kinetic_energy (torch.Tensor): Kinetic energy associated with the innermost NH thermostats.

        Returns:
            torch.Tensor: Scaling factor applied to the system momenta.

        References
        ----------
        .. [#nhc_thermostat2] Martyna, Tuckerman, Tobias, Klein:
           Explicit reversible integrators for extended systems dynamics.
           Molecular Physics, 87(5), 1117-1157. 1996.
        """
        # Compute forces on first thermostat
        self.forces[..., 0] = (
            kinetic_energy - self.degrees_of_freedom * self.kb_temperature
        ) / self.masses[..., 0]

        scaling_factor = 1.0
        for _ in range(self.multi_step):
            for idx_ys in range(self.integration_order):
                time_step = self.time_step[idx_ys]

                # Update velocities of outermost bath
                self.velocities[..., -1] += 0.25 * self.forces[..., -1] * time_step

                # Update the velocities moving through the beads of the chain
                for chain in range(self.chain_length - 2, -1, -1):
                    coeff = torch.exp(
                        -0.125 * time_step * self.velocities[..., chain + 1]
                    )
                    self.velocities[..., chain] = (
                        self.velocities[..., chain] * coeff ** 2
                        + 0.25 * self.forces[..., chain] * coeff * time_step
                    )

                # Accumulate velocity scaling
                scaling_factor *= torch.exp(-0.5 * time_step * self.velocities[..., 0])
                # Update forces of innermost thermostat
                self.forces[..., 0] = (
                    scaling_factor * scaling_factor * kinetic_energy
                    - self.degrees_of_freedom * self.kb_temperature
                ) / self.masses[..., 0]

                # Update thermostat positions
                # TODO: Only required if one is interested in the conserved
                #  quanity of the NHC.
                self.positions += 0.5 * self.velocities * time_step

                # Update the thermostat velocities
                for chain in range(self.chain_length - 1):
                    coeff = torch.exp(
                        -0.125 * time_step * self.velocities[..., chain + 1]
                    )
                    self.velocities[..., chain] = (
                        self.velocities[..., chain] * coeff ** 2
                        + 0.25 * self.forces[..., chain] * coeff * time_step
                    )
                    self.forces[..., chain + 1] = (
                        self.masses[..., chain] * self.velocities[..., chain] ** 2
                        - self.kb_temperature
                    ) / self.masses[..., chain + 1]

                # Update velocities of outermost thermostat
                self.velocities[..., -1] += 0.25 * self.forces[..., -1] * time_step

        return scaling_factor

    def _compute_kinetic_energy(self, momenta, masses):
        """
        Routine for computing the kinetic energy of the innermost NH thermostats based on the momenta and masses of the
        simulated systems.

        Args:
            momenta (torch.Tensor): Momenta of the simulated system.
            masses (torch.Tensor): Masses of the simulated system.

        Returns:
            torch.Tensor: Kinetic energy associated with the innermost NH thermostats. These are summed over the
                          corresponding degrees of freedom, depending on whether a massive NHC is used.

        """
        # Compute the kinetic energy (factor of 1/2 can be removed, as it
        # cancels with a times 2)
        # TODO: Is no problem, as NM transformation never mixes atom dimension
        #  which carries the masses.
        kinetic_energy = momenta ** 2 / masses
        if self.massive:
            return kinetic_energy
        else:
            return torch.sum(
                torch.sum(kinetic_energy, 3, keepdim=True), 2, keepdim=True
            )

    def _apply_thermostat(self, simulator):
        """
        Propagate the NHC thermostat, compute the corresponding scaling factor and apply it to the momenta of the
        system. If a normal mode transformer is provided, this is done in the normal model representation of the ring
        polymer.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Get current momenta
        momenta = simulator.system.momenta

        # Apply transformation
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.beads2normal(momenta)

        kinetic_energy = self._compute_kinetic_energy(momenta, simulator.system.masses)

        scaling_factor = self._propagate_thermostat(kinetic_energy)
        momenta = momenta * scaling_factor

        # self.compute_conserved(simulator.system)

        # Apply transformation if requested
        if self.nm_transformation is not None:
            momenta = self.nm_transformation.normal2beads(momenta)

        simulator.system.momenta = momenta

    def compute_conserved(self, system):
        conserved = (
            system.kinetic_energy[..., None, None]
            + 0.5 * torch.sum(self.velocities ** 2 * self.masses, 4)
            + system.energies[..., None, None]
            + self.degrees_of_freedom * self.kb_temperature * self.positions[..., 0]
            + self.kb_temperature * torch.sum(self.positions[..., 1:], 4)
        )
        print(conserved)

    @property
    def state_dict(self):
        state_dict = {
            "chain_length": self.chain_length,
            "massive": self.massive,
            "frequency": self.frequency,
            "kb_temperature": self.kb_temperature,
            "degrees_of_freedom": self.degrees_of_freedom,
            "masses": self.masses,
            "velocities": self.velocities,
            "forces": self.forces,
            "positions": self.positions,
            "time_step": self.time_step,
            "temperature_bath": self.temperature_bath,
            "n_replicas": self.n_replicas,
            "multi_step": self.multi_step,
            "integration_order": self.integration_order,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.chain_length = state_dict["chain_length"]
        self.massive = state_dict["massive"]
        self.frequency = state_dict["frequency"]
        self.kb_temperature = state_dict["kb_temperature"]
        self.degrees_of_freedom = state_dict["degrees_of_freedom"]
        self.masses = state_dict["masses"]
        self.velocities = state_dict["velocities"]
        self.forces = state_dict["forces"]
        self.positions = state_dict["positions"]
        self.time_step = state_dict["time_step"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.n_replicas = state_dict["n_replicas"]
        self.multi_step = state_dict["multi_step"]
        self.integration_order = state_dict["integration_order"]

        self.initialized = True


class NHCRingPolymerThermostat(NHCThermostat):
    """
    Nose-Hoover chain thermostat for ring polymer molecular dynamics simulations as e.g. described in
    [#stochastic_thermostats4]_. This is based on the massive setting of the standard NHC thermostat but operates in
    the normal mode representation and uses specially initialized thermostat masses.

    Args:
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        chain_length (int): Number of Nose-Hoover thermostats applied in the chain.
        local (bool): If set to true, an individual thermostat is applied to each degree of freedom in the system.
                        Can e.g. be used for thermostatting (default=False).
        nm_transformation (schnetpack.md.utils.NormalModeTransformer): Module used to transform between beads and normal
                                                                   model representation in ring polymer dynamics.
        multi_step (int): Number of steps used for integrating the NH equations of motion (default=2)
        integration_order (int): Order of the Yoshida-Suzuki integrator used for propagating the thermostat (default=3).


    References
    ----------
    .. [#stochastic_thermostats4] Ceriotti, Parrinello, Markland, Manolopoulos:
       Efficient stochastic thermostatting of path integral molecular dynamics.
       The Journal of Chemical Physics, 133 (12), 124104. 2010.
    """

    def __init__(
        self,
        temperature_bath,
        time_constant,
        chain_length=3,
        local=True,
        nm_transformation=NormalModeTransformer,
        multi_step=2,
        integration_order=3,
    ):
        super(NHCRingPolymerThermostat, self).__init__(
            temperature_bath,
            time_constant,
            chain_length=chain_length,
            massive=True,
            nm_transformation=nm_transformation,
            multi_step=multi_step,
            integration_order=integration_order,
        )
        self.local = local

    def _init_masses(self, state_dimension, simulator):
        """
        Initialize masses according to the normal mode frequencies of the ring polymer system.

        Args:
            state_dimension (tuple): Size of the thermostat states. This is used to differentiate between the massive
                                     and the standard algorithm
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on the
                                                                 time step, system, etc.
        """
        # Multiply factor by number of replicas
        self.kb_temperature = self.kb_temperature * self.n_replicas

        # Initialize masses with the frequencies of the ring polymer
        polymer_frequencies = simulator.integrator.omega_normal
        # 0.5 comes from Ceriotti paper, check
        polymer_frequencies[0] = 0.5 * self.frequency

        # Assume standard massive Nose-Hoover and initialize accordingly
        self.masses = torch.ones(state_dimension, device=self.device)
        self.masses *= (
            self.kb_temperature / polymer_frequencies[:, None, None, None, None] ** 2
        )

        # If a global thermostat is requested, we assign masses of 3N to
        # the first link in the chain on the centroid
        if not self.local:
            self.masses[0, :, :, :, 0] *= (
                3 * simulator.system.n_atoms.float()[:, None, None]
            )
            # Degrees of freedom also need to be adapted
            self.degrees_of_freedom[0, :, :, :] *= (
                3 * simulator.system.n_atoms.float()[:, None, None]
            )

    def _compute_kinetic_energy(self, momenta, masses):
        """
        Routine for computing the kinetic energies of the innermost NH thermostats based on the masses and momenta of
        the ring polymer in normal mode representation.

        Args:
            momenta (torch.Tensor): Normal mode momenta of the simulated system.
            masses (torch.Tensor): Masses of the simulated system.

        Returns:
            torch.Tensor: Kinetic energy of the innermost NH thermostats.
        """
        kinetic_energy = momenta ** 2 / masses

        # In case of a global NHC for RPMD, use the whole centroid kinetic
        # energy and broadcast it
        if not self.local:
            kinetic_energy_centroid = torch.sum(
                torch.sum(kinetic_energy[0, ...], 2, keepdim=True), 1, keepdim=True
            )
            kinetic_energy[0, ...] = kinetic_energy_centroid

        return kinetic_energy

    @property
    def state_dict(self):
        state_dict = {
            "chain_length": self.chain_length,
            "massive": self.massive,
            "frequency": self.frequency,
            "kb_temperature": self.kb_temperature,
            "degrees_of_freedom": self.degrees_of_freedom,
            "masses": self.masses,
            "velocities": self.velocities,
            "forces": self.forces,
            "positions": self.positions,
            "time_step": self.time_step,
            "temperature_bath": self.temperature_bath,
            "n_replicas": self.n_replicas,
            "multi_step": self.multi_step,
            "integration_order": self.integration_order,
            "local": self.local,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.chain_length = state_dict["chain_length"]
        self.massive = state_dict["massive"]
        self.frequency = state_dict["frequency"]
        self.kb_temperature = state_dict["kb_temperature"]
        self.degrees_of_freedom = state_dict["degrees_of_freedom"]
        self.masses = state_dict["masses"]
        self.velocities = state_dict["velocities"]
        self.forces = state_dict["forces"]
        self.positions = state_dict["positions"]
        self.time_step = state_dict["time_step"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.n_replicas = state_dict["n_replicas"]
        self.multi_step = state_dict["multi_step"]
        self.integration_order = state_dict["integration_order"]
        self.local = state_dict["local"]

        self.initialized = True



import argparse
import logging

from schnetpack.md.parsers.md_setup import MDSimulation

try:
    import oyaml as yaml
except ImportError:
    import yaml


def read_options(yamlpath):
    with open(yamlpath, "r") as tf:
        tradoffs = yaml.safe_load(tf)

    logging.info("Read options from {:s}.".format(yamlpath))
    return tradoffs


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("md_input")
    args = parser.parse_args()

    config = read_options(args.md_input)

    md = MDSimulation(config)
    md.save_config()
    md.run()



from .basic_hooks import *
from .thermostats import *
from .barostats import *
from .sampling import *
from .logging_hooks import *

# class MTKBarostat(BarostatHook):
#    def __init__(self, target_pressure, time_constant, temperature_bath, detach=True):
#        super(MTKBarostat, self).__init__(
#           target_pressure=target_pressure,
#          temperature_bath=temperature_bath,
#          detach=detach,
#      )
#      self.frequency = 1.0 / (time_constant * MDUnits.fs2atu)
#      self.inv_kb_temperature = 1.0 / (self.temperature_bath * MDUnits.kB)

#    def _init_barostat(self, simulator):
# TODO: perform checks here?

#        # Get scaling factors based on the number of atoms
#        n_atoms = simulator.system.n_atoms[None, :, None, None].float()
#        self.inv_sqrt_dof = 1.0 / torch.sqrt(3.0 * n_atoms + 1.0)
#        self.inv_dof = 1.0 / n_atoms
#        self.weight = self.frequency * self.inv_sqrt_dof
#
#        # Thermostat variable, this will be used for scaling the positions and momenta
#        self.zeta = torch.zeros(
#            self.n_replicas, self.n_molecules, 1, 1, device=self.device
#        )
#
#    def _apply_barostat(self, simulator):
#        # TODO: check dimensions and pressure units
#
#        # 1) Update barostat variable
#        # pressure_factor = 3.0 * simulator.system.volume * (
#        #        simulator.system.compute_pressure(kinetic_component=True) - self.target_pressure
#        # ) + 2.0 * self.inv_dof * simulator.system.kinetic_energy
#
#        # The pressure factor is 3*V * (Pint - Pext) + 2*Ekin / N
#        # Since Pint also contains the kinetic component it computes as
#        #   Pint = 1/(3V) * 2 * Ekin + Pint'
#        # This means, the expression can be rewritten as
#        #   Pfact = 3*V*(Pint' - Pext) + 2*(1+1/N) * Ekin
#        # Saving some computations
#        pressure_factor = (
#                3.0
#                * simulator.system.volume
#                * (simulator.system.compute_pressure() - self.target_pressure)
#                + 2.0 * (1 + self.inv_dof) * simulator.system.kinetic_energy
#        )
#
#        self.zeta = (
#                self.zeta
#                + 0.5
#                * self.time_step
#                * self.weight
#                * self.inv_kb_temperature
#                * pressure_factor
#        )
#
#        # 2) Scale positions and cell
#        scaling = torch.exp(self.time_step * self.weight * self.zeta)
#
#        simulator.system.positions = simulator.system.positions * scaling
#        simulator.system.cells = simulator.system.cells * scaling
#
#        # 3) Scale momenta
#        scaling = torch.exp(
#            -self.time_step * self.weight * (1 + self.inv_dof) * self.zeta
#        )
#        simulator.system.momenta = simulator.system.momenta * scaling
#
#        # 4) Second update of barostat variable based on new momenta and positions
#        pressure_factor = (
#                3.0
#                * simulator.system.volume
#                * (simulator.system.compute_pressure() - self.target_pressure)
#                + 2.0 * (1 + self.inv_dof) * simulator.system.kinetic_energy
#        )
#
#        self.zeta = (
#                self.zeta
#                + 0.5
#                * self.time_step
#                * self.weight
#                * self.inv_kb_temperature
#                * pressure_factor
#       )
#
#        if self.detach:
#            simulator.system.cells = simulator.system.cells.detach()
#            simulator.system.positions = simulator.system.positions.detach()
#            simulator.system.momenta = simulator.system.momenta.detach()
#            self.zeta = self.zeta.detach()
